from datetime import datetime
from typing import Literal, Optional

import numpy as np
from pynwb import NWBFile
from pynwb.base import DynamicTable

from .mock_ttl_signals import generate_mock_ttl_signal
from ...basetemporalalignmentinterface import BaseTemporalAlignmentInterface
from ...datainterfaces import SpikeGLXNIDQInterface
from ...datainterfaces.ecephys.baserecordingextractorinterface import (
    BaseRecordingExtractorInterface,
)
from ...datainterfaces.ecephys.basesortingextractorinterface import (
    BaseSortingExtractorInterface,
)
from ...datainterfaces.ophys.baseimagingextractorinterface import (
    BaseImagingExtractorInterface,
)
from ...datainterfaces.ophys.basesegmentationextractorinterface import (
    BaseSegmentationExtractorInterface,
)
from ...utils import ArrayType, get_schema_from_method_signature


class MockBehaviorEventInterface(BaseTemporalAlignmentInterface):
    """
    A mock behavior event interface for testing purposes.
    """

    @classmethod
    def get_source_schema(cls) -> dict:
        source_schema = get_schema_from_method_signature(method=cls.__init__, exclude=["event_times"])
        source_schema["additionalProperties"] = True
        return source_schema

    def __init__(self, event_times: Optional[ArrayType] = None):
        """
        Define event times for some behavior.

        Parameters
        ----------
        event_times : list of floats, optional
            The event times to set as timestamps for this interface.
            The default is the array [1.2, 2.3, 3.4] for similarity to the timescale of the MockSpikeGLXNIDQInterface.
        """
        event_times = event_times or [1.2, 2.3, 3.4]
        self.event_times = np.array(event_times)
        self.original_event_times = np.array(event_times)  # Make a copy of the initial loaded timestamps

    def get_original_timestamps(self) -> np.ndarray:
        return self.original_event_times

    def get_timestamps(self) -> np.ndarray:
        return self.event_times

    def set_aligned_timestamps(self, aligned_timestamps: np.ndarray):
        self.event_times = aligned_timestamps

    def add_to_nwbfile(self, nwbfile: NWBFile, metadata: dict):
        table = DynamicTable(name="BehaviorEvents", description="Times of various classified behaviors.")
        table.add_column(name="event_time", description="Time of each event.")
        for timestamp in self.get_timestamps():  # adding data by column gives error
            table.add_row(event_time=timestamp)
        nwbfile.add_acquisition(table)


class MockSpikeGLXNIDQInterface(SpikeGLXNIDQInterface):
    """
    A mock SpikeGLX interface for testing purposes.
    """

    ExtractorName = "NumpyRecording"

    @classmethod
    def get_source_schema(cls) -> dict:
        source_schema = get_schema_from_method_signature(method=cls.__init__, exclude=["ttl_times"])
        source_schema["additionalProperties"] = True
        return source_schema

    def __init__(
        self, signal_duration: float = 7.0, ttl_times: Optional[list[list[float]]] = None, ttl_duration: float = 1.0
    ):
        """
        Define a mock SpikeGLXNIDQInterface by overriding the recording extractor to be a mock TTL signal.

        Parameters
        ----------
        signal_duration : float, default: 7.0
            The number of seconds to simulate.
        ttl_times : list of lists of floats, optional
            The times within the `signal_duration` to trigger the TTL pulse for each channel.
            The outer list is over channels, while each inner list is the set of TTL times for each specific channel.
            The default generates 8 channels with periodic on/off cycle (which start in the 'off' state)
            each of which is of length `ttl_duration` with a 0.1 second offset per channel.
        ttl_duration : float, default: 1.0
            How long the TTL pulses stays in the 'on' state when triggered, in seconds.
        """
        from spikeinterface.extractors import NumpyRecording

        if ttl_times is None:
            # Begin in 'off' state
            number_of_periods = int(np.ceil((signal_duration - ttl_duration) / (ttl_duration * 2)))
            default_periodic_ttl_times = [ttl_duration * (1 + 2 * period) for period in range(number_of_periods)]
            ttl_times = [[ttl_time + 0.1 * channel for ttl_time in default_periodic_ttl_times] for channel in range(8)]
        number_of_channels = len(ttl_times)
        channel_ids = [f"nidq#XA{channel_index}" for channel_index in range(number_of_channels)]  # NIDQ channel IDs
        channel_groups = ["NIDQChannelGroup"] * number_of_channels

        sampling_frequency = 25_000.0  # NIDQ sampling rate
        number_of_frames = int(signal_duration * sampling_frequency)
        traces = np.empty(shape=(number_of_frames, number_of_channels), dtype="int16")
        for channel_index in range(number_of_channels):
            traces[:, channel_index] = generate_mock_ttl_signal(
                signal_duration=signal_duration,
                ttl_times=ttl_times[channel_index],
                ttl_duration=ttl_duration,
                sampling_frequency_hz=sampling_frequency,
            )

        self.recording_extractor = NumpyRecording(
            traces_list=traces, sampling_frequency=sampling_frequency, channel_ids=channel_ids
        )
        # NIDQ channel gains
        self.recording_extractor.set_channel_gains(gains=[61.03515625] * self.recording_extractor.get_num_channels())
        self.recording_extractor.set_property(key="group_name", values=channel_groups)

        # Minimal meta so `get_metadata` works similarly to real NIDQ header
        self.meta = {"acqMnMaXaDw": "0,0,8,1", "fileCreateTime": "2020-11-03T10:35:10", "niDev1ProductName": "PCI-6259"}
        self.subset_channels = None
        self.verbose = None
        self.es_key = "ElectricalSeriesNIDQ"


class MockRecordingInterface(BaseRecordingExtractorInterface):
    """An interface with a spikeinterface recording object for testing purposes."""

    ExtractorModuleName = "spikeinterface.core.generate"
    ExtractorName = "generate_recording"

    def __init__(
        self,
        num_channels: int = 4,
        sampling_frequency: float = 30_000.0,
        durations: tuple[float] = (1.0,),
        seed: int = 0,
        verbose: bool = True,
        es_key: str = "ElectricalSeries",
    ):
        super().__init__(
            num_channels=num_channels,
            sampling_frequency=sampling_frequency,
            durations=durations,
            seed=seed,
            verbose=verbose,
            es_key=es_key,
        )

    def get_metadata(self) -> dict:
        metadata = super().get_metadata()
        session_start_time = datetime.now().astimezone()
        metadata["NWBFile"]["session_start_time"] = session_start_time
        return metadata


class MockSortingInterface(BaseSortingExtractorInterface):
    """A mock sorting extractor interface for generating synthetic sorting data."""

    # TODO: Implement this class with the lazy generator once is merged
    # https://github.com/SpikeInterface/spikeinterface/pull/2227

    ExtractorModuleName = "spikeinterface.core.generate"
    ExtractorName = "generate_sorting"

    def __init__(
        self,
        num_units: int = 4,
        sampling_frequency: float = 30_000.0,
        durations: tuple[float] = (1.0,),
        seed: int = 0,
        verbose: bool = True,
    ):
        """
        Parameters
        ----------
        num_units : int, optional
            Number of units to generate, by default 4.
        sampling_frequency : float, optional
            Sampling frequency of the generated data in Hz, by default 30,000.0 Hz.
        durations : tuple of float, optional
            Durations of the segments in seconds, by default (1.0,).
        seed : int, optional
            Seed for the random number generator, by default 0.
        verbose : bool, optional
            Control whether to display verbose messages during writing, by default True.

        """

        super().__init__(
            num_units=num_units,
            sampling_frequency=sampling_frequency,
            durations=durations,
            seed=seed,
            verbose=verbose,
        )

    def get_metadata(self) -> dict:  # noqa D102
        metadata = super().get_metadata()
        session_start_time = datetime.now().astimezone()
        metadata["NWBFile"]["session_start_time"] = session_start_time
        return metadata


class MockImagingInterface(BaseImagingExtractorInterface):
    """
    A mock imaging interface for testing purposes.
    """

    ExtractorModuleName = "roiextractors.testing"
    ExtractorName = "generate_dummy_imaging_extractor"

    def __init__(
        self,
        num_frames: int = 30,
        num_rows: int = 10,
        num_columns: int = 10,
        sampling_frequency: float = 30,
        dtype: str = "uint16",
        verbose: bool = False,
        seed: int = 0,
        photon_series_type: Literal["OnePhotonSeries", "TwoPhotonSeries"] = "TwoPhotonSeries",
    ):
        """
        Parameters
        ----------
        num_frames : int, optional
            The number of frames in the mock imaging data, by default 30.
        num_rows : int, optional
            The number of rows (height) in each frame of the mock imaging data, by default 10.
        num_columns : int, optional
            The number of columns (width) in each frame of the mock imaging data, by default 10.
        sampling_frequency : float, optional
            The sampling frequency of the mock imaging data in Hz, by default 30.
        dtype : str, optional
            The data type of the generated imaging data (e.g., 'uint16'), by default 'uint16'.
        seed : int, optional
            Random seed for reproducibility, by default 0.
        photon_series_type : Literal["OnePhotonSeries", "TwoPhotonSeries"], optional
            The type of photon series for the mock imaging data, either "OnePhotonSeries" or
            "TwoPhotonSeries", by default "TwoPhotonSeries".
        verbose : bool, default False
            controls verbosity
        """

        self.seed = seed
        super().__init__(
            num_frames=num_frames,
            num_rows=num_rows,
            num_columns=num_columns,
            sampling_frequency=sampling_frequency,
            dtype=dtype,
            verbose=verbose,
        )

        self.verbose = verbose
        self.photon_series_type = photon_series_type

    def get_metadata(self, photon_series_type: Optional[Literal["OnePhotonSeries", "TwoPhotonSeries"]] = None) -> dict:
        session_start_time = datetime.now().astimezone()
        metadata = super().get_metadata(photon_series_type=photon_series_type)
        metadata["NWBFile"]["session_start_time"] = session_start_time
        return metadata


class MockSegmentationInterface(BaseSegmentationExtractorInterface):
    """A mock segmentation interface for testing purposes."""

    ExtractorModuleName = "roiextractors.testing"
    ExtractorName = "generate_dummy_segmentation_extractor"

    def __init__(
        self,
        num_rois: int = 10,
        num_frames: int = 30,
        num_rows: int = 25,
        num_columns: int = 25,
        sampling_frequency: float = 30.0,
        has_summary_images: bool = True,
        has_raw_signal: bool = True,
        has_dff_signal: bool = True,
        has_deconvolved_signal: bool = True,
        has_neuropil_signal: bool = True,
        seed: int = 0,
        verbose: bool = False,
    ):
        """
        Parameters
        ----------
        num_rois : int, optional
            number of regions of interest, by default 10.
        num_frames : int, optional
            description, by default 30.
        num_rows : int, optional
            number of rows in the hypothetical video from which the data was extracted, by default 25.
        num_columns : int, optional
            number of columns in the hypothetical video from which the data was extracted, by default 25.
        sampling_frequency : float, optional
            sampling frequency of the hypothetical video from which the data was extracted, by default 30.0.
        has_summary_images : bool, optional
            whether the dummy segmentation extractor has summary images or not (mean and correlation).
        has_raw_signal : bool, optional
            whether a raw fluorescence signal is desired in the object, by default True.
        has_dff_signal : bool, optional
            whether a relative (df/f) fluorescence signal is desired in the object, by default True.
        has_deconvolved_signal : bool, optional
            whether a deconvolved signal is desired in the object, by default True.
        has_neuropil_signal : bool, optional
            whether a neuropil signal is desired in the object, by default True.
        seed: int, default 0
            seed for the random number generator, by default 0
        verbose : bool, optional
            controls verbosity, by default False.
        """

        super().__init__(
            num_rois=num_rois,
            num_frames=num_frames,
            num_rows=num_rows,
            num_columns=num_columns,
            sampling_frequency=sampling_frequency,
            has_summary_images=has_summary_images,
            has_raw_signal=has_raw_signal,
            has_dff_signal=has_dff_signal,
            has_deconvolved_signal=has_deconvolved_signal,
            has_neuropil_signal=has_neuropil_signal,
            verbose=verbose,
        )

    def get_metadata(self) -> dict:
        session_start_time = datetime.now().astimezone()
        metadata = super().get_metadata()
        metadata["NWBFile"]["session_start_time"] = session_start_time
        return metadata



from typing import Optional

import numpy as np
from pydantic import FilePath, validate_call
from pynwb.behavior import BehavioralEpochs, IntervalSeries
from pynwb.file import NWBFile

from neuroconv.basetemporalalignmentinterface import BaseTemporalAlignmentInterface
from neuroconv.tools import get_package, nwb_helpers
from neuroconv.utils import DeepDict

from .medpc_helpers import read_medpc_file


class MedPCInterface(BaseTemporalAlignmentInterface):
    """
    Data Interface for MedPC output files.

    The output files from MedPC are raw text files that contain behavioral data from the operant box sessions such as
    lever presses, reward port entries, nose pokes, etc. The output text files format this data into a series of
    colon-separated variables that are either single-line (for metadata) or multi-line (for arrays). The multi-line
    variables keep a colon-separated index of the array every 5 elements.  For example, a single variable might look like::

        Start Date: 11/09/18

    while a multi-line variable might look like::

        A:
            0:      175.150      270.750      762.050      762.900     1042.600
            5:     1567.800     1774.950     2448.450     2454.050     2552.800
            10:     2620.550     2726.250

    Different sessions are usually separated by a blank line or two.

    This data is parsed by the MedPCInterface and added to the NWBFile as Events and IntervalSeries objects in the
    behavior module.
    """

    keywords = ("behavior",)
    display_name = "MedPC"
    info = "Interface for handling MedPC output files."
    associated_suffixes = (".txt",)

    @validate_call
    def __init__(
        self,
        file_path: FilePath,
        session_conditions: dict,
        start_variable: str,
        metadata_medpc_name_to_info_dict: dict,
        aligned_timestamp_names: Optional[list[str]] = None,
        verbose: bool = True,
    ):
        """
        Initialize MedpcInterface.

        Parameters
        ----------
        file_path : FilePath
            Path to the MedPC file.
        session_conditions : dict
            The conditions that define the session. The keys are the names of the single-line variables (ex. 'Start Date')
            and the values are the values of those variables for the desired session (ex. '11/09/18').
        start_variable : str
            The name of the variable that starts the session (ex. 'Start Date').
        metadata_medpc_name_to_info_dict : dict
            A dictionary mapping the names of the desired variables in the MedPC file
            to an info dictionary with the names of the variables in the metadata and whether or not they are arrays.
            ex. {"Start Date": {"name": "start_date", "is_array": False}}
        aligned_timestamp_names : list, optional
            The names of the variables that are externally aligned timestamps,
            which should be retrieved from self.timestamps_dict instead of the MedPC output file.
        verbose : bool, optional
            Whether to print verbose output, by default True
        """
        if aligned_timestamp_names is None:
            aligned_timestamp_names = []
        super().__init__(
            file_path=file_path,
            session_conditions=session_conditions,
            start_variable=start_variable,
            metadata_medpc_name_to_info_dict=metadata_medpc_name_to_info_dict,
            aligned_timestamp_names=aligned_timestamp_names,
            verbose=verbose,
        )
        self.timestamps_dict = {}

    def get_metadata(self) -> DeepDict:
        metadata = super().get_metadata()
        session_dict = read_medpc_file(
            file_path=self.source_data["file_path"],
            medpc_name_to_info_dict=self.source_data["metadata_medpc_name_to_info_dict"],
            session_conditions=self.source_data["session_conditions"],
            start_variable=self.source_data["start_variable"],
        )
        for k, v in session_dict.items():
            metadata["MedPC"][k] = v

        return metadata

    def get_metadata_schema(self) -> dict:
        metadata_schema = super().get_metadata_schema()
        medpc_name_to_info_dict = self.source_data["metadata_medpc_name_to_info_dict"]
        metadata_schema["properties"]["MedPC"] = {
            "type": "object",
            "properties": {info_dict["name"]: {"type": "string"} for info_dict in medpc_name_to_info_dict.values()},
        }
        return metadata_schema

    def get_original_timestamps(self, medpc_name_to_info_dict: dict) -> dict[str, np.ndarray]:
        """
        Retrieve the original unaltered timestamps dictionary for the data in this interface.

        This function retrieves the data on-demand by re-reading the medpc file.

        Parameters
        ----------
        medpc_name_to_info_dict : dict
            A dictionary mapping the names of the desired variables in the MedPC file
            to an info dictionary with the names of the variables in the metadata and whether or not they are arrays.
            ex. {"A": {"name": "left_nose_poke_times", "is_array": True}}

        Returns
        -------
        timestamps_dict: dict
            A dictionary mapping the names of the variables to the original medpc timestamps.
        """
        timestamps_dict = read_medpc_file(
            file_path=self.source_data["file_path"],
            medpc_name_to_info_dict=medpc_name_to_info_dict,
            session_conditions=self.source_data["session_conditions"],
            start_variable=self.source_data["start_variable"],
        )
        return timestamps_dict

    def get_timestamps(self) -> dict[str, np.ndarray]:
        """
        Retrieve the timestamps dictionary for the data in this interface.

        Returns
        -------
        timestamps_dict: dict
            A dictionary mapping the names of the variables to the timestamps.
        """
        return self.timestamps_dict

    def set_aligned_timestamps(self, aligned_timestamps_dict: dict[str, np.ndarray]) -> None:
        """
        Replace all timestamps for this interface with those aligned to the common session start time.

        Must be in units seconds relative to the common 'session_start_time'.

        Parameters
        ----------
        aligned_timestamps_dict : dict
            A dictionary mapping the names of the variables to the synchronized timestamps for data in this interface.
        """
        self.timestamps_dict = aligned_timestamps_dict

    def set_aligned_starting_time(self, aligned_starting_time: float, medpc_name_to_info_dict: dict) -> None:
        """
        Align the starting time for this interface relative to the common session start time.

        Must be in units seconds relative to the common 'session_start_time'.

        Parameters
        ----------
        aligned_starting_time : float
            The starting time for all temporal data in this interface.
        medpc_name_to_info_dict : dict
            A dictionary mapping the names of the desired variables in the MedPC file
            to an info dictionary with the names of the variables in the metadata and whether or not they are arrays.
            ex. {"A": {"name": "left_nose_poke_times", "is_array": True}}
        """
        original_timestamps_dict = self.get_original_timestamps(medpc_name_to_info_dict=medpc_name_to_info_dict)
        aligned_timestamps_dict = {}
        for name, original_timestamps in original_timestamps_dict.items():
            aligned_timestamps_dict[name] = original_timestamps + aligned_starting_time
        self.set_aligned_timestamps(aligned_timestamps_dict=aligned_timestamps_dict)

    def add_to_nwbfile(
        self,
        nwbfile: NWBFile,
        metadata: dict,
    ) -> None:
        ndx_events = get_package(package_name="ndx_events", installation_instructions="pip install ndx-events")
        medpc_name_to_info_dict = metadata["MedPC"].get("medpc_name_to_info_dict", None)
        assert medpc_name_to_info_dict is not None, "medpc_name_to_info_dict must be provided in metadata"
        info_name_to_medpc_name = {
            info_dict["name"]: medpc_name for medpc_name, info_dict in medpc_name_to_info_dict.items()
        }
        for name in self.source_data["aligned_timestamp_names"]:
            medpc_name = info_name_to_medpc_name[name]
            medpc_name_to_info_dict.pop(medpc_name)
        session_dict = read_medpc_file(
            file_path=self.source_data["file_path"],
            medpc_name_to_info_dict=medpc_name_to_info_dict,
            session_conditions=self.source_data["session_conditions"],
            start_variable=self.source_data["start_variable"],
        )
        aligned_timestamps_dict = self.get_timestamps()
        for name, aligned_timestamps in aligned_timestamps_dict.items():
            session_dict[name] = aligned_timestamps

        # Add behavior data to nwbfile
        module_name = metadata["MedPC"].get("module_name", "behavior")
        module_description = metadata["MedPC"].get("module_description", "Behavioral data from MedPC output files.")
        behavior_module = nwb_helpers.get_module(
            nwbfile=nwbfile,
            name=module_name,
            description=module_description,
        )

        event_dicts = metadata["MedPC"].get("Events", [])
        for event_dict in event_dicts:
            name = event_dict["name"]
            description = event_dict["description"]
            event_data = session_dict[name]
            if len(event_data) > 0:
                event = ndx_events.Events(
                    name=name,
                    description=description,
                    timestamps=event_data,
                )
                behavior_module.add(event)
        interval_dicts = metadata["MedPC"].get("IntervalSeries", [])
        for interval_dict in interval_dicts:
            name = interval_dict["name"]
            description = interval_dict["description"]
            onset_name = interval_dict["onset_name"]
            duration_name = interval_dict["duration_name"]
            onset_data = session_dict[onset_name]
            duration_data = session_dict[duration_name]
            if len(onset_data) == 0:
                continue
            assert not len(duration_data) == 0, f"Duration data for {name} is empty!"

            interval_times, data = [], []
            for onset_time, duration in zip(onset_data, duration_data):
                interval_times.append(onset_time)
                data.append(1)
                interval_times.append(onset_time + duration)
                data.append(-1)
            interval = IntervalSeries(
                name=name,
                description=description,
                timestamps=interval_times,
                data=data,
            )
            behavioral_epochs = BehavioralEpochs(name="behavioral_epochs")
            behavioral_epochs.add_interval_series(interval)
            behavior_module.add(behavioral_epochs)



import numpy as np


def generate_mock_probe(num_channels: int, num_shanks: int = 3):
    """
    Generate a mock probe with specified number of channels and shanks.

    Parameters:
        num_channels (int): The number of channels in the probe.
        num_shanks (int, optional): The number of shanks in the probe. Defaults to 3.

    Returns:
        pi.Probe: The generated mock probe.

    """
    import probeinterface as pi

    # The shank ids will be 0, 0, 0, ..., 1, 1, 1, ..., 2, 2, 2, ...
    shank_ids: list[int] = []
    positions = np.zeros((num_channels, 2))
    # ceil division
    channels_per_shank = (num_channels + num_shanks - 1) // num_shanks
    for i in range(num_shanks):
        # x0, y0 is the position of the first electrode in the shank
        x0 = 0
        y0 = i * 200
        for j in range(channels_per_shank):
            if len(shank_ids) == num_channels:
                break
            shank_ids.append(i)
            x = x0 + j * 10
            y = y0 + (j % 2) * 10
            positions[len(shank_ids) - 1] = x, y
    probe = pi.Probe(ndim=2, si_units="um")
    probe.set_contacts(positions=positions, shapes="circle", shape_params={"radius": 5})
    probe.set_device_channel_indices(np.arange(num_channels))
    probe.set_shank_ids(shank_ids)
    return probe



import collections.abc
import json
import warnings
from collections import defaultdict
from copy import deepcopy
from ctypes import Union
from pathlib import Path
from typing import Any, Optional, Union

import numpy as np
import yaml
from pydantic import FilePath


class _NoDatesSafeLoader(yaml.SafeLoader):
    """Custom override of yaml Loader class for datetime considerations."""

    @classmethod
    def remove_implicit_resolver(cls, tag_to_remove):
        """
        Remove implicit resolvers for a particular tag.

        Takes care not to modify resolvers in super classes.
        Solution taken from https://stackoverflow.com/a/37958106/11483674
        We want to load datetimes as strings, not dates, because we go on to serialise as jsonwhich doesn't have the
        advanced types of yaml, and leads to incompatibilities down the track.
        """
        if "yaml_implicit_resolvers" not in cls.__dict__:
            cls.yaml_implicit_resolvers = cls.yaml_implicit_resolvers.copy()
        for first_letter, mappings in cls.yaml_implicit_resolvers.items():
            cls.yaml_implicit_resolvers[first_letter] = [
                (tag, regexp) for tag, regexp in mappings if tag != tag_to_remove
            ]


_NoDatesSafeLoader.remove_implicit_resolver("tag:yaml.org,2002:timestamp")


def load_dict_from_file(file_path: FilePath) -> dict:
    """Safely load metadata from .yml or .json files."""
    file_path = Path(file_path)
    assert file_path.is_file(), f"{file_path} is not a file."
    assert file_path.suffix in (".json", ".yml", ".yaml"), f"{file_path} is not a valid yaml or .json file."

    if file_path.suffix in (".yml", ".yaml"):
        with open(file=file_path, mode="r") as stream:
            dictionary = yaml.load(stream=stream, Loader=_NoDatesSafeLoader)
    elif file_path.suffix == ".json":
        with open(file=file_path, mode="r") as fp:
            dictionary = json.load(fp=fp)
    return dictionary


def exist_dict_in_list(d, ls):
    """Check if an identical dictionary exists in the list."""
    return any([d == i for i in ls])


def append_replace_dict_in_list(ls, d, compare_key, list_dict_deep_update: bool = True, remove_repeats: bool = True):
    """
    Update the list ls with the dict d.

    Cases:

    1.  If d is a dict and ls a list of dicts and ints/str, then for a given compare key, if for any element of ls
        (which is a dict) say: ``ls[3][compare_key] == d[compare_key]``, then it will dict_deep_update these instead of
        appending d to list ls. Only if compare_key is not present in any of dicts in the list ls, then d is simply
        appended to ls.
    2.  If ``d`` is of immutable types like str, int etc., the ls is either appended with ``d`` or not.
        This depends on the value of ``remove_repeats``. If ``remove_repeats`` is ``False``, then ls is always appended with d.
        If ``remove_repeats`` is ``True``, then if value d is present then it is not appended else it is.

    Parameters
    ----------
    ls: list
        list of a dicts or int/str or a combination. This is the object to update
    d: list/str/int
        this is the object from which ls is updated.
    compare_key: str
        name of the key for which to check the presence of dicts in ls which need dict_deep_update
    list_dict_deep_update: bool
        whether to update a dict in ls with compare_key present OR simply replace it.
    remove_repeats: bool
        keep repeated values in the updated ls
    Returns
    -------
    ls: list
        updated list
    """
    if not isinstance(ls, list):
        return d
    if isinstance(d, collections.abc.Mapping):
        indxs = np.where(
            [d.get(compare_key, None) == i[compare_key] for i in ls if isinstance(i, collections.abc.Mapping)]
        )[0]
        if len(indxs) > 0:
            for idx in indxs:
                if list_dict_deep_update:
                    ls[idx] = dict_deep_update(ls[idx], d)
                else:
                    ls[idx] = d
        else:
            ls.append(d)
    elif not (d in ls and remove_repeats):
        ls.append(d)
    return ls


def dict_deep_update(
    d: collections.abc.Mapping,
    u: collections.abc.Mapping,
    append_list: bool = True,
    remove_repeats: bool = True,
    copy: bool = True,
    compare_key: str = "name",
    list_dict_deep_update: bool = True,
) -> collections.abc.Mapping:
    """
    Perform an update to all nested keys of dictionary d(input) from dictionary u(updating dict).

    Parameters
    ----------
    d: dict
        dictionary to update
    u: dict
        dictionary to update from
    append_list: bool
        if the item to update is a list, whether to append the lists or replace the list in d
        e.g. d = dict(key1=[1,2,3]), u = dict(key1=[3,4,5]).
        If True then updated dictionary d=dict(key1=[1,2,3,4,5]) else d=dict(key1=[3,4,5])
    remove_repeats: bool
        for updating list in d[key] with list in u[key]: if true then remove repeats: list(set(ls))
    copy: bool
        whether to deepcopy the input dict d
    compare_key: str
        the key that is used to compare dicts (and perform update op) and update d[key] when it is a list if dicts.
        example::

            d = {
                [
                    {"name": "timeseries1", "desc": "desc1 of d", "starting_time": 0.0},
                    {"name": "timeseries2", "desc": "desc2"},
                ]
            }
            u = [{"name": "timeseries1", "desc": "desc2 of u", "unit": "n.a."}]
            # if compare_key='name' output is below
            output = [
                {"name": "timeseries1", "desc": "desc2 of u", "starting_time": 0.0, "unit": "n.a."},
                {"name": "timeseries2", "desc": "desc2"},
            ]
            # else the output is:
            # dict with the same key will be updated instead of being appended to the list
            output = [
                {"name": "timeseries1", "desc": "desc1 of d", "starting_time": 0.0},
                {"name": "timeseries2", "desc": "desc2"},
                {"name": "timeseries1", "desc": "desc2 of u", "unit": "n.a."},
            ]

    list_dict_deep_update: bool
        for back compatibility, if False, this would work as before:
        example: if True then for the compare_key example, the output would be::

            output = [
                {"name": "timeseries1", "desc": "desc2 of u", "starting_time": 0.0, "unit": "n.a."},
                {"name": "timeseries2", "desc": "desc2"},
            ]
            # if False:
            output = [
                {"name": "timeseries1", "desc": "desc2 of u", "starting_time": 0.0},
                {"name": "timeseries2", "desc": "desc2"},
            ]  # unit key is absent since it is a replacement
    Returns
    -------
    d: dict
        return the updated dictionary
    """

    dict_to_update, dict_with_update_values = d, u
    if not isinstance(dict_to_update, collections.abc.Mapping):
        warnings.warn("input to update should be a dict, returning output")
        return dict_with_update_values

    if copy:
        dict_to_update = deepcopy(dict_to_update)

    for key_to_update, update_values in dict_with_update_values.items():
        # Update with a dict like object is recursive until an empty dict is found.
        if isinstance(update_values, collections.abc.Mapping):
            sub_dict_to_update = dict_to_update.get(key_to_update, dict())
            sub_dict_with_update_values = update_values
            dict_to_update[key_to_update] = dict_deep_update(
                sub_dict_to_update, sub_dict_with_update_values, append_list=append_list, remove_repeats=remove_repeats
            )
        # Update with list calls the append_replace_dict_in_list function
        elif append_list and isinstance(update_values, list):
            for value in update_values:
                dict_or_list_of_dicts = dict_to_update.get(key_to_update, [])
                dict_to_update[key_to_update] = append_replace_dict_in_list(
                    dict_or_list_of_dicts, value, compare_key, list_dict_deep_update, remove_repeats
                )
        # Update with something else
        else:
            dict_to_update[key_to_update] = update_values

    return dict_to_update


class DeepDict(defaultdict):
    """A defaultdict of defaultdicts"""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(lambda: DeepDict(), *args, **kwargs)
        for key, value in self.items():
            if isinstance(value, dict):
                self[key] = DeepDict(value)

    def deep_update(self, other: Optional[Union[dict, "DeepDict"]] = None, **kwargs) -> None:
        for key, value in (other or kwargs).items():
            if key in self and isinstance(self[key], dict) and isinstance(value, dict):
                self[key].deep_update(value)
            else:
                self[key] = value

    def to_dict(self) -> dict:
        """Turn a DeepDict into a normal dictionary"""

        def _to_dict(d: Union[dict, "DeepDict"]) -> dict:
            return {key: _to_dict(value) for key, value in d.items()} if isinstance(d, dict) else d

        return _to_dict(self)

    def __deepcopy__(self, memodict={}):
        """

        Parameters
        ----------
        memodict: dict
            unused

        Returns
        -------
        DeepDict

        """
        return DeepDict(deepcopy(self.to_dict()))

    def __repr__(self) -> str:
        return "DeepDict: " + dict.__repr__(self.to_dict())



from copy import deepcopy
from typing import Literal, Optional, Union

import numpy as np
from pynwb import NWBFile
from pynwb.device import Device
from pynwb.ecephys import ElectrodeGroup

from .baserecordingextractorinterface import BaseRecordingExtractorInterface
from ...baseextractorinterface import BaseExtractorInterface
from ...utils import DeepDict, get_base_schema, get_schema_from_hdmf_class


class BaseSortingExtractorInterface(BaseExtractorInterface):
    """Primary class for all SortingExtractor interfaces."""

    keywords = ("extracellular electrophysiology", "spike sorting")

    ExtractorModuleName = "spikeinterface.extractors"

    def __init__(self, verbose=True, **source_data):
        super().__init__(**source_data)
        self.sorting_extractor = self.get_extractor()(**source_data)
        self.verbose = verbose
        self._number_of_segments = self.sorting_extractor.get_num_segments()

    def get_metadata_schema(self) -> dict:
        """Compile metadata schema for the RecordingExtractor."""

        # Initiate Ecephys metadata
        metadata_schema = super().get_metadata_schema()
        metadata_schema["properties"]["Ecephys"] = get_base_schema(tag="Ecephys")
        metadata_schema["properties"]["Ecephys"]["required"] = []
        metadata_schema["properties"]["Ecephys"]["properties"] = dict(
            Device=dict(type="array", minItems=1, items={"$ref": "#/properties/Ecephys/definitions/Device"}),
            ElectrodeGroup=dict(
                type="array", minItems=1, items={"$ref": "#/properties/Ecephys/definitions/ElectrodeGroup"}
            ),
            Electrodes=dict(
                type="array",
                minItems=0,
                renderForm=False,
                items={"$ref": "#/properties/Ecephys/definitions/Electrodes"},
            ),
            UnitProperties=dict(
                type="array",
                minItems=0,
                renderForm=False,
                items={"$ref": "#/properties/Ecephys/definitions/UnitProperties"},
            ),
        )

        # Schema definition for arrays
        metadata_schema["properties"]["Ecephys"]["definitions"] = dict(
            Device=get_schema_from_hdmf_class(Device),
            ElectrodeGroup=get_schema_from_hdmf_class(ElectrodeGroup),
            Electrodes=dict(
                type="object",
                additionalProperties=False,
                required=["name"],
                properties=dict(
                    name=dict(type="string", description="name of this electrodes column"),
                    description=dict(type="string", description="description of this electrodes column"),
                ),
            ),
            UnitProperties=dict(
                type="object",
                additionalProperties=False,
                required=["name"],
                properties=dict(
                    name=dict(type="string", description="name of this units column"),
                    description=dict(type="string", description="description of this units column"),
                ),
            ),
        )
        return metadata_schema

    def register_recording(self, recording_interface: BaseRecordingExtractorInterface):
        self.sorting_extractor.register_recording(recording=recording_interface.recording_extractor)

    def get_original_timestamps(self) -> np.ndarray:
        raise NotImplementedError(
            "Unable to fetch original timestamps for a SortingInterface since it relies upon an attached recording."
        )

    def get_timestamps(self) -> Union[np.ndarray, list[np.ndarray]]:
        if not self.sorting_extractor.has_recording():
            raise NotImplementedError(
                "In order to align timestamps for a SortingInterface, it must have a recording "
                "object attached to it! Please attach one by calling `.register_recording(recording_interface=...)`."
            )
        if self._number_of_segments == 1:
            return self.sorting_extractor._recording.get_times()
        else:
            return [
                self.sorting_extractor._recording.get_times(segment_index=segment_index)
                for segment_index in range(self._number_of_segments)
            ]

    def set_aligned_timestamps(self, aligned_timestamps: np.ndarray):
        """
        Replace all timestamps for the attached interface with those aligned to the common session start time.

        Must be in units seconds relative to the common 'session_start_time'.
        Must have a single-segment RecordingInterface attached; call `.register_recording(recording_interface=...)` to accomplish this.

        When a SortingInterface has a recording attached, it infers the timing via the frame indices of the
        timestamps from the corresponding recording segment. This method aligns the timestamps of that recording
        so that the SortingExtractor can automatically infer the timing from the frames.

        Parameters
        ----------
        aligned_timestamps : numpy.ndarray or list of numpy.ndarray
            The synchronized timestamps for data in this interface.
            If there is more than one segment in the sorting/recording pair, then
        """
        if not self.sorting_extractor.has_recording():
            raise NotImplementedError(
                "In order to align timestamps for a SortingInterface, it must have a recording "
                "object attached to it! Please attach one by calling `.register_recording(recording_interface=...)`."
            )
        assert (
            self._number_of_segments == 1
        ), "This recording has multiple segments; please use 'set_aligned_segment_timestamps' instead."

        if self._number_of_segments == 1:
            self.sorting_extractor._recording.set_times(times=aligned_timestamps, with_warning=False)
        else:
            assert isinstance(
                aligned_timestamps, list
            ), "Recording has multiple segment! Please pass a list of timestamps to align each segment."
            assert (
                len(aligned_timestamps) == self._number_of_segments
            ), f"The number of timestamp vectors ({len(aligned_timestamps)}) does not match the number of segments ({self._number_of_segments})!"

            for segment_index in range(self._number_of_segments):
                self.sorting_extractor._recording.set_times(
                    times=aligned_timestamps[segment_index],
                    segment_index=segment_index,
                    with_warning=False,
                )

    def set_aligned_segment_timestamps(self, aligned_segment_timestamps: list[np.ndarray]):
        """
        Replace all timestamps for all segments in this interface with those aligned to the common session start time.

        Must be in units seconds relative to the common 'session_start_time'.
        Must have a multi-segment RecordingInterface attached by calling `.register_recording(recording_interface=...)`.

        Parameters
        ----------
        aligned_segment_timestamps : list of numpy.ndarray
            The synchronized timestamps for segment of data in this interface.
        """
        if not self.sorting_extractor.has_recording():
            raise NotImplementedError(
                "In order to align timestamps for a SortingInterface, it must have a recording "
                "object attached to it! Please attach one by calling `.register_recording(recording_interface=...)`."
            )
        assert isinstance(
            aligned_segment_timestamps, list
        ), "Recording has multiple segment! Please pass a list of timestamps to align each segment."
        assert (
            len(aligned_segment_timestamps) == self._number_of_segments
        ), f"The number of timestamp vectors ({len(aligned_segment_timestamps)}) does not match the number of segments ({self._number_of_segments})!"

        for segment_index in range(self._number_of_segments):
            self.sorting_extractor._recording.set_times(
                times=aligned_segment_timestamps[segment_index],
                segment_index=segment_index,
                with_warning=False,
            )

    def set_aligned_starting_time(self, aligned_starting_time: float):
        if self.sorting_extractor.has_recording():
            if self._number_of_segments == 1:
                self.set_aligned_timestamps(aligned_timestamps=self.get_timestamps() + aligned_starting_time)
            else:
                self.set_aligned_segment_timestamps(
                    [segment_timestamps + aligned_starting_time for segment_timestamps in self.get_timestamps()]
                )
        else:
            for sorting_segment in self.sorting_extractor._sorting_segments:
                if sorting_segment._t_start is None:
                    sorting_segment._t_start = aligned_starting_time
                else:
                    sorting_segment._t_start += aligned_starting_time

    def set_aligned_segment_starting_times(self, aligned_segment_starting_times: list[float]):
        """
        Align the starting time for each segment in this interface relative to the common session start time.

        Must be in units seconds relative to the common 'session_start_time'.

        Parameters
        ----------
        aligned_segment_starting_times : list of floats
            The starting time for each segment of data in this interface.
        """
        assert len(aligned_segment_starting_times) == self._number_of_segments, (
            f"The length of the starting_times ({len(aligned_segment_starting_times)}) does not match the number "
            f"of segments ({self._number_of_segments})!"
        )

        if self.sorting_extractor.has_recording():
            if self._number_of_segments == 1:
                self.set_aligned_starting_time(aligned_starting_time=aligned_segment_starting_times[0])
            else:
                aligned_segment_timestamps = [
                    segment_timestamps + aligned_segment_starting_time
                    for segment_timestamps, aligned_segment_starting_time in zip(
                        self.get_timestamps(), aligned_segment_starting_times
                    )
                ]
                self.set_aligned_segment_timestamps(aligned_segment_timestamps=aligned_segment_timestamps)
        else:
            for sorting_segment, aligned_segment_starting_time in zip(
                self.sorting_extractor._sorting_segments, aligned_segment_starting_times
            ):
                sorting_segment._t_start = aligned_segment_starting_time

    def subset_sorting(self):
        max_min_spike_time = max(
            [
                min(x)
                for y in self.sorting_extractor.get_unit_ids()
                for x in [self.sorting_extractor.get_unit_spike_train(y)]
                if any(x)
            ]
        )
        end_frame = int(1.1 * max_min_spike_time)
        if self.sorting_extractor.has_recording():
            end_frame = min(end_frame, self.sorting_extractor._recording.get_total_samples())
        stub_sorting_extractor = self.sorting_extractor.frame_slice(start_frame=0, end_frame=end_frame)
        return stub_sorting_extractor

    def add_channel_metadata_to_nwb(self, nwbfile: NWBFile, metadata: Optional[DeepDict] = None):
        """
        Add channel metadata to an NWBFile object using information extracted from a SortingExtractor and
        optional metadata.

        This function attempts to add devices, electrode groups, and electrodes to the NWBFile. If a recording is
        associated with the SortingExtractor, it is used for metadata addition. Otherwise, it attempts to create a dummy
        NumpyRecording based on the provided metadata. If neither is available, the function warns the user and skips the
        metadata addition.

        Parameters
        ----------
        nwbfile : NWBFile
            The NWBFile object to which the metadata is added.
        metadata : Optional[DeepDict]
            Optional metadata to use for the addition of electrode-related data. If it's provided, it should contain an
            "Ecephys" field with a nested "Electrodes" field.

        Returns
        -------
        None

        Raises
        ------
        Warning
            If there's no recording in the sorting extractor and no electrodes metadata in the provided metadata, a warning
            is raised and the function returns None.

        Notes
        -----
        This function adds metadata to the `nwbfile` in-place, meaning the `nwbfile` object is modified directly.
        """
        from ...tools.spikeinterface import (
            add_devices_to_nwbfile,
            add_electrode_groups_to_nwbfile,
            add_electrodes_to_nwbfile,
        )

        if hasattr(self, "generate_recording_with_channel_metadata"):
            recording = self.generate_recording_with_channel_metadata()

            add_devices_to_nwbfile(nwbfile=nwbfile, metadata=metadata)
            add_electrode_groups_to_nwbfile(recording=recording, nwbfile=nwbfile, metadata=metadata)
            add_electrodes_to_nwbfile(recording=recording, nwbfile=nwbfile, metadata=metadata)

    def add_to_nwbfile(
        self,
        nwbfile: NWBFile,
        metadata: Optional[DeepDict] = None,
        stub_test: bool = False,
        write_ecephys_metadata: bool = False,
        write_as: Literal["units", "processing"] = "units",
        units_name: str = "units",
        units_description: str = "Autogenerated by neuroconv.",
    ):
        """
        Primary function for converting the data in a SortingExtractor to NWB format.

        Parameters
        ----------
        nwbfile : NWBFile
            Fill the relevant fields within the NWBFile object.
        metadata : DeepDict
            Information for constructing the NWB file (optional) and units table descriptions.
            Should be of the format::

                metadata["Ecephys"]["UnitProperties"] = dict(name=my_name, description=my_description)
        stub_test : bool, default: False
            If True, will truncate the data to run the conversion faster and take up less memory.
        write_ecephys_metadata : bool, default: False
            Write electrode information contained in the metadata.
        write_as : {'units', 'processing'}
            How to save the units table in the nwb file. Options:
            - 'units' will save it to the official NWBFile.Units position; recommended only for the final form of the data.
            - 'processing' will save it to the processing module to serve as a historical provenance for the official table.
        units_name : str, default: 'units'
            The name of the units table. If write_as=='units', then units_name must also be 'units'.
        units_description : str, default: 'Autogenerated by neuroconv.'
        """
        from ...tools.spikeinterface import add_sorting_to_nwbfile

        metadata_copy = deepcopy(metadata)
        if write_ecephys_metadata:
            self.add_channel_metadata_to_nwb(nwbfile=nwbfile, metadata=metadata_copy)

        if stub_test:
            sorting_extractor = self.subset_sorting()
        else:
            sorting_extractor = self.sorting_extractor

        property_descriptions = dict()
        for metadata_column in metadata_copy["Ecephys"].get("UnitProperties", []):
            property_descriptions.update({metadata_column["name"]: metadata_column["description"]})
            for unit_id in sorting_extractor.get_unit_ids():
                # Special condition for wrapping electrode group pointers to actual object ids rather than string names
                if metadata_column["name"] == "electrode_group" and nwbfile.electrode_groups:
                    value = nwbfile.electrode_groups[
                        self.sorting_extractor.get_unit_property(unit_id=unit_id, property_name="electrode_group")
                    ]
                    sorting_extractor.set_unit_property(
                        unit_id=unit_id,
                        property_name=metadata_column["name"],
                        value=value,
                    )

        add_sorting_to_nwbfile(
            sorting_extractor,
            nwbfile=nwbfile,
            property_descriptions=property_descriptions,
            write_as=write_as,
            units_name=units_name,
            units_description=units_description,
        )



from pydantic import DirectoryPath

from ..basesortingextractorinterface import BaseSortingExtractorInterface


class KiloSortSortingInterface(BaseSortingExtractorInterface):
    """Primary data interface class for converting a KiloSortingExtractor from spikeinterface."""

    display_name = "KiloSort Sorting"
    associated_suffixes = (".npy",)
    info = "Interface for KiloSort sorting data."

    @classmethod
    def get_source_schema(cls) -> dict:
        source_schema = super().get_source_schema()
        source_schema["properties"]["folder_path"][
            "description"
        ] = "Path to the output Phy folder (containing the params.py)"
        return source_schema

    def __init__(
        self,
        folder_path: DirectoryPath,
        keep_good_only: bool = False,
        verbose: bool = True,
    ):
        """
        Load and prepare sorting data for kilosort

        Parameters
        ----------
        folder_path: str or Path
            Path to the output Phy folder (containing the params.py)
        keep_good_only: bool, default: False
            If True, only Kilosort-labeled 'good' units are returned
        verbose: bool, default: True
        """
        super().__init__(folder_path=folder_path, keep_good_only=keep_good_only, verbose=verbose)

    def get_metadata(self):
        metadata = super().get_metadata()
        # See Kilosort save_to_phy() docstring for more info on these fields: https://github.com/MouseLand/Kilosort/blob/main/kilosort/io.py
        # Or see phy documentation: https://github.com/cortex-lab/phy/blob/master/phy/apps/base.py
        metadata["Ecephys"]["UnitProperties"] = [
            dict(name="n_spikes", description="Number of spikes recorded from each unit."),
            dict(name="fr", description="Average firing rate of each unit."),
            dict(name="depth", description="Estimated depth of each unit in micrometers."),
            dict(name="Amplitude", description="Per-template amplitudes, computed as the L2 norm of the template."),
            dict(
                name="ContamPct",
                description="Contamination rate for each template, computed as fraction of refractory period violations relative to expectation based on a Poisson process.",
            ),
            dict(
                name="KSLabel",
                description="Label indicating whether each template is 'mua' (multi-unit activity) or 'good' (refractory).",
            ),
            dict(name="original_cluster_id", description="Original cluster ID assigned by Kilosort."),
            dict(
                name="amp",
                description="For every template, the maximum amplitude of the template waveforms across all channels.",
            ),
            dict(name="ch", description="The channel label of the best channel, as defined by the user."),
            dict(name="sh", description="The shank label of the best channel."),
        ]
        return metadata



from pathlib import Path

from pydantic import FilePath
from pynwb.ecephys import ElectricalSeries

from ..baserecordingextractorinterface import BaseRecordingExtractorInterface
from ....utils import get_schema_from_hdmf_class


class IntanRecordingInterface(BaseRecordingExtractorInterface):
    """
    Primary data interface class for converting Intan data using the

    :py:class:`~spikeinterface.extractors.IntanRecordingExtractor`.
    """

    display_name = "Intan Recording"
    associated_suffixes = (".rhd", ".rhs")
    info = "Interface for Intan recording data."
    stream_id = "0"  # This are the amplifier channels, corresponding to the stream_name 'RHD2000 amplifier channel'

    @classmethod
    def get_source_schema(cls) -> dict:
        source_schema = super().get_source_schema()
        source_schema["properties"]["file_path"]["description"] = "Path to either a .rhd or a .rhs file"
        return source_schema

    def __init__(
        self,
        file_path: FilePath,
        verbose: bool = True,
        es_key: str = "ElectricalSeries",
        ignore_integrity_checks: bool = False,
    ):
        """
        Load and prepare raw data and corresponding metadata from the Intan format (.rhd or .rhs files).

        Parameters
        ----------
        file_path : FilePathType
            Path to either a rhd or a rhs file

        verbose : bool, default: True
            Verbose
        es_key : str, default: "ElectricalSeries"
        ignore_integrity_checks, bool, default: False.
            If True, data that violates integrity assumptions will be loaded. At the moment the only integrity
            check performed is that timestamps are continuous. If False, an error will be raised if the check fails.
        """

        self.file_path = Path(file_path)

        init_kwargs = dict(
            file_path=self.file_path,
            stream_id=self.stream_id,
            verbose=verbose,
            es_key=es_key,
            all_annotations=True,
            ignore_integrity_checks=ignore_integrity_checks,
        )

        super().__init__(**init_kwargs)

    def get_metadata_schema(self) -> dict:
        metadata_schema = super().get_metadata_schema()
        metadata_schema["properties"]["Ecephys"]["properties"].update(
            ElectricalSeriesRaw=get_schema_from_hdmf_class(ElectricalSeries)
        )
        return metadata_schema

    def get_metadata(self) -> dict:
        metadata = super().get_metadata()
        ecephys_metadata = metadata["Ecephys"]

        # Add device

        system = self.file_path.suffix  # .rhd or .rhs
        device_description = {".rhd": "RHD Recording System", ".rhs": "RHS Stim/Recording System"}[system]

        intan_device = dict(
            name="Intan",
            description=device_description,
            manufacturer="Intan",
        )
        device_list = [intan_device]
        ecephys_metadata.update(Device=device_list)

        electrode_group_metadata = ecephys_metadata["ElectrodeGroup"]
        electrode_group_metadata[0]["device"] = intan_device["name"]

        # Add electrodes and electrode groups
        ecephys_metadata.update(
            ElectricalSeriesRaw=dict(name="ElectricalSeriesRaw", description="Raw acquisition traces."),
        )

        return metadata



from .basedatainterface import BaseDataInterface
from .baseextractorinterface import BaseExtractorInterface
from .basetemporalalignmentinterface import BaseTemporalAlignmentInterface
from .nwbconverter import ConverterPipe, NWBConverter
from .tools import get_format_summaries
from .tools.yaml_conversion_specification import run_conversion_from_yaml



from typing import Literal, Optional

from dateutil.parser import parse
from pydantic import DirectoryPath

from ..baseimagingextractorinterface import BaseImagingExtractorInterface
from ....utils.dict import DeepDict


class BrukerTiffMultiPlaneImagingInterface(BaseImagingExtractorInterface):
    """Interface for Bruker multi-plane TIFF files using BrukerTiffMultiPlaneImagingExtractor from roiextractors."""

    display_name = "Bruker TIFF Imaging (single channel, multiple planes)"
    associated_suffixes = (".ome", ".tif", ".xml", ".env")
    info = "Interface for a single channel of multi-plane Bruker TIFF imaging data."

    @classmethod
    def get_source_schema(cls) -> dict:
        source_schema = super().get_source_schema()
        source_schema["properties"]["folder_path"][
            "description"
        ] = "The folder that contains the Bruker TIF image files (.ome.tif) and configuration files (.xml, .env)."
        return source_schema

    @classmethod
    def get_streams(
        cls,
        folder_path: DirectoryPath,
        plane_separation_type: Literal["contiguous", "disjoint"] = None,
    ) -> dict:
        from roiextractors import BrukerTiffMultiPlaneImagingExtractor

        streams = BrukerTiffMultiPlaneImagingExtractor.get_streams(folder_path=folder_path)
        for channel_stream_name in streams["channel_streams"]:
            if plane_separation_type == "contiguous":
                streams["plane_streams"].update(
                    {channel_stream_name: [streams["plane_streams"][channel_stream_name][0]]}
                )
        return streams

    def __init__(
        self,
        folder_path: DirectoryPath,
        stream_name: Optional[str] = None,
        verbose: bool = True,
    ):
        """
        Initialize reading of TIFF files.

        Parameters
        ----------
        folder_path : DirectoryPath
            The path to the folder that contains the Bruker TIF image files (.ome.tif) and configuration files (.xml, .env).
        stream_name : str, optional
            The name of the recording stream (e.g. 'Ch2').
        verbose : bool, default: True
        """
        self.folder_path = folder_path
        super().__init__(
            folder_path=folder_path,
            stream_name=stream_name,
            verbose=verbose,
        )
        self._stream_name = self.imaging_extractor.stream_name.replace("_", "")
        self._image_size = self.imaging_extractor.get_image_size()

    def _determine_position_current(self) -> list[float]:
        """
        Returns y, x, and z position values. The unit of values is in the microscope reference frame.
        """
        from roiextractors.extractors.tiffimagingextractors.brukertiffimagingextractor import (
            _parse_xml,
        )

        streams = self.get_streams(folder_path=self.folder_path)
        channel_stream_name = streams["channel_streams"][0]
        plane_streams_per_channel = streams["plane_streams"][channel_stream_name]

        # general positionCurrent
        position_values = []
        xml_root_element = _parse_xml(folder_path=self.folder_path)
        default_position_element = xml_root_element.find(".//PVStateValue[@key='positionCurrent']")

        for index_value in ["YAxis", "XAxis"]:
            position_sub_indexed_values = default_position_element.find(f"./SubindexedValues[@index='{index_value}']")
            for position_sub_indexed_value in position_sub_indexed_values:
                position_values.append(float(position_sub_indexed_value.attrib["value"]))

        z_plane_values = []
        for plane_stream in plane_streams_per_channel:
            frames_per_stream = [
                frame
                for frame in xml_root_element.findall(".//Frame")
                for file in frame.findall("File")
                if plane_stream in file.attrib["filename"]
            ]

            # The frames for each plane will have the same positionCurrent values
            position_element = frames_per_stream[0].find(".//PVStateValue[@key='positionCurrent']")
            default_z_position_values = default_position_element.find("./SubindexedValues[@index='ZAxis']")
            z_positions = []
            for z_sub_indexed_value in default_z_position_values:
                z_value = float(z_sub_indexed_value.attrib["value"])
                z_positions.append(z_value)

            z_position_values = position_element.find("./SubindexedValues[@index='ZAxis']")
            for z_device_ind, z_position_value in enumerate(z_position_values):
                z_value = float(z_position_value.attrib["value"])
                # find the changing z position value
                if z_positions[z_device_ind] != z_value:
                    z_plane_values.append(z_value)

        # difference between start position and end position of the z scan
        if len(z_plane_values) > 1:
            z_value = abs(z_plane_values[0] - z_plane_values[-1])
        position_values.append(z_value)
        return position_values

    def get_metadata(self) -> DeepDict:
        metadata = super().get_metadata()

        xml_metadata = self.imaging_extractor.xml_metadata
        session_start_time = parse(xml_metadata["date"])
        metadata["NWBFile"].update(session_start_time=session_start_time)

        description = f"Version {xml_metadata['version']}"
        device_name = "BrukerFluorescenceMicroscope"
        metadata["Ophys"]["Device"][0].update(
            name=device_name,
            description=description,
        )

        imaging_plane_metadata = metadata["Ophys"]["ImagingPlane"][0]
        imaging_plane_metadata.update(
            device=device_name,
            imaging_rate=self.imaging_extractor.get_sampling_frequency(),
        )
        two_photon_series_metadata = metadata["Ophys"]["TwoPhotonSeries"][0]
        two_photon_series_metadata.update(
            description="The volumetric imaging data acquired from the Bruker Two-Photon Microscope.",
            scan_line_rate=1 / float(xml_metadata["scanLinePeriod"]),
        )

        streams = self.get_streams(folder_path=self.folder_path, plane_separation_type="contiguous")
        if len(streams["channel_streams"]) > 1:
            imaging_plane_name = f"ImagingPlane{self._stream_name}"
            imaging_plane_metadata.update(name=imaging_plane_name)
            two_photon_series_metadata.update(
                name=f"TwoPhotonSeries{self._stream_name}",
                imaging_plane=imaging_plane_name,
            )

        microns_per_pixel = xml_metadata["micronsPerPixel"]
        x_position_in_meters = float(microns_per_pixel[0]["XAxis"]) / 1e6
        y_position_in_meters = float(microns_per_pixel[1]["YAxis"]) / 1e6

        origin_coords = self._determine_position_current()
        z_plane_current_position_in_meters = abs(origin_coords[-1]) / 1e6
        grid_spacing = [y_position_in_meters, x_position_in_meters, z_plane_current_position_in_meters]
        field_of_view = [
            y_position_in_meters * self._image_size[1],
            x_position_in_meters * self._image_size[0],
            z_plane_current_position_in_meters,
        ]

        imaging_plane_metadata.update(
            grid_spacing=grid_spacing,
            origin_coords=origin_coords,
            description="The imaging plane origin_coords units are in the microscope reference frame.",
        )

        two_photon_series_metadata.update(field_of_view=field_of_view)

        return metadata


class BrukerTiffSinglePlaneImagingInterface(BaseImagingExtractorInterface):
    """Data Interface for BrukerTiffSinglePlaneImagingExtractor."""

    display_name = "Bruker TIFF Imaging (single channel, single plane)"
    associated_suffixes = BrukerTiffMultiPlaneImagingInterface.associated_suffixes
    info = "Interface for handling a single channel and a single plane of Bruker TIFF imaging data."

    @classmethod
    def get_source_schema(cls) -> dict:
        source_schema = super().get_source_schema()
        source_schema["properties"]["folder_path"][
            "description"
        ] = "The folder containing the Bruker TIF image files and configuration files."
        return source_schema

    @classmethod
    def get_streams(cls, folder_path: DirectoryPath) -> dict:
        from roiextractors import BrukerTiffMultiPlaneImagingExtractor

        streams = BrukerTiffMultiPlaneImagingExtractor.get_streams(folder_path=folder_path)
        return streams

    def __init__(
        self,
        folder_path: DirectoryPath,
        stream_name: Optional[str] = None,
        verbose: bool = True,
    ):
        """
        Initialize reading of TIFF files.

        Parameters
        ----------
        folder_path : DirectoryPath
            The path to the folder that contains the Bruker TIF image files (.ome.tif) and configuration files (.xml, .env).
        stream_name : str, optional
            The name of the recording stream (e.g. 'Ch2').
        verbose : bool, default: True
        """
        super().__init__(
            folder_path=folder_path,
            stream_name=stream_name,
            verbose=verbose,
        )
        self.folder_path = folder_path
        self._stream_name = self.imaging_extractor.stream_name.replace("_", "")
        self._image_size = self.imaging_extractor.get_image_size()

    def _determine_position_current(self) -> list[float]:
        """
        Returns y, x, and z position values. The unit of values is in the microscope reference frame.
        """
        stream_name = self.imaging_extractor.stream_name
        frames_per_stream = [
            frame
            for frame in self.imaging_extractor._xml_root.findall(".//Frame")
            for file in frame.findall("File")
            if stream_name in file.attrib["filename"]
        ]

        # general positionCurrent
        position_values = []
        default_position_element = self.imaging_extractor._xml_root.find(".//PVStateValue[@key='positionCurrent']")

        for index_value in ["YAxis", "XAxis"]:
            position_sub_indexed_values = default_position_element.find(f"./SubindexedValues[@index='{index_value}']")
            for position_sub_indexed_value in position_sub_indexed_values:
                position_values.append(float(position_sub_indexed_value.attrib["value"]))

        # The frames for each plane will have the same positionCurrent values
        position_element = frames_per_stream[0].find(".//PVStateValue[@key='positionCurrent']")
        if not position_element:
            return position_values

        default_z_position_values = default_position_element.find("./SubindexedValues[@index='ZAxis']")
        z_positions = []
        for z_sub_indexed_value in default_z_position_values:
            z_positions.append(float(z_sub_indexed_value.attrib["value"]))

        z_position_values = position_element.find("./SubindexedValues[@index='ZAxis']")
        for z_device_ind, z_position_value in enumerate(z_position_values):
            z_value = float(z_position_value.attrib["value"])
            # find the changing z position value
            if z_positions[z_device_ind] != z_value:
                position_values.append(z_value)

        return position_values

    def get_metadata(self) -> DeepDict:
        metadata = super().get_metadata()

        xml_metadata = self.imaging_extractor.xml_metadata
        session_start_time = parse(xml_metadata["date"])
        metadata["NWBFile"].update(session_start_time=session_start_time)

        description = f"Version {xml_metadata['version']}"
        device_name = "BrukerFluorescenceMicroscope"
        metadata["Ophys"]["Device"][0].update(
            name=device_name,
            description=description,
        )

        imaging_plane_metadata = metadata["Ophys"]["ImagingPlane"][0]
        imaging_plane_metadata.update(
            device=device_name,
            imaging_rate=self.imaging_extractor.get_sampling_frequency(),
        )
        two_photon_series_metadata = metadata["Ophys"]["TwoPhotonSeries"][0]

        two_photon_series_metadata.update(
            description="Imaging data acquired from the Bruker Two-Photon Microscope.",
            scan_line_rate=1 / float(xml_metadata["scanLinePeriod"]),
        )

        streams = self.get_streams(folder_path=self.folder_path)
        if len(streams["channel_streams"]) > 1 or len(streams["plane_streams"]):
            imaging_plane_name = f"ImagingPlane{self._stream_name}"
            imaging_plane_metadata.update(name=imaging_plane_name)
            two_photon_series_metadata.update(
                name=f"TwoPhotonSeries{self._stream_name}",
                imaging_plane=imaging_plane_name,
            )

        microns_per_pixel = xml_metadata["micronsPerPixel"]
        x_position_in_meters = float(microns_per_pixel[0]["XAxis"]) / 1e6
        y_position_in_meters = float(microns_per_pixel[1]["YAxis"]) / 1e6
        grid_spacing = [y_position_in_meters, x_position_in_meters]
        origin_coords = self._determine_position_current()
        field_of_view = [
            y_position_in_meters * self._image_size[1],
            x_position_in_meters * self._image_size[0],
        ]

        if len(streams["plane_streams"]) and len(origin_coords) == 3:
            z_plane_current_position_in_meters = abs(origin_coords[-1]) / 1e6
            grid_spacing.append(z_plane_current_position_in_meters)
            field_of_view.append(z_plane_current_position_in_meters)

        imaging_plane_metadata.update(
            grid_spacing=grid_spacing,
            origin_coords=origin_coords,
            description="The imaging plane origin_coords units are in the microscope reference frame.",
        )
        two_photon_series_metadata.update(field_of_view=field_of_view)

        return metadata



"""Collection of helper functions related to configuration of datasets dependent on backend."""

import importlib
from typing import Union

from hdmf.common import Data
from pynwb import NWBFile, TimeSeries

from ._configuration_models._hdf5_backend import HDF5BackendConfiguration
from ._configuration_models._zarr_backend import ZarrBackendConfiguration
from ..importing import is_package_installed


def configure_backend(
    nwbfile: NWBFile, backend_configuration: Union[HDF5BackendConfiguration, ZarrBackendConfiguration]
) -> None:
    """
    Configure all datasets specified in the `backend_configuration` with their appropriate DataIO and options.

    Parameters
    ----------
    nwbfile : pynwb.NWBFile
        The in-memory pynwb.NWBFile object to configure.
    backend_configuration : HDF5BackendConfiguration or ZarrBackendConfiguration
        The configuration model to use when configuring the datasets for this backend.
    """
    is_ndx_events_installed = is_package_installed(package_name="ndx_events")
    ndx_events = importlib.import_module("ndx_events") if is_ndx_events_installed else None

    # A remapping of the object IDs in the backend configuration might necessary
    locations_to_remap = backend_configuration.find_locations_requiring_remapping(nwbfile=nwbfile)
    if any(locations_to_remap):
        backend_configuration = backend_configuration.build_remapped_backend(locations_to_remap=locations_to_remap)

    # Set all DataIO based on the configuration
    data_io_class = backend_configuration.data_io_class
    for dataset_configuration in backend_configuration.dataset_configurations.values():
        object_id = dataset_configuration.object_id
        dataset_name = dataset_configuration.dataset_name
        data_io_kwargs = dataset_configuration.get_data_io_kwargs()

        # TODO: update buffer shape in iterator, if present

        neurodata_object = nwbfile.objects[object_id]
        is_dataset_linked = isinstance(neurodata_object.fields.get(dataset_name), TimeSeries)

        # Table columns
        if isinstance(neurodata_object, Data):
            neurodata_object.set_data_io(data_io_class=data_io_class, data_io_kwargs=data_io_kwargs)
        # TimeSeries data or timestamps
        elif isinstance(neurodata_object, TimeSeries) and not is_dataset_linked:
            neurodata_object.set_data_io(
                dataset_name=dataset_name, data_io_class=data_io_class, data_io_kwargs=data_io_kwargs
            )
        # Special ndx-events v0.2.0 types
        elif is_ndx_events_installed and isinstance(neurodata_object, ndx_events.Events):
            neurodata_object.set_data_io(
                dataset_name=dataset_name, data_io_class=data_io_class, data_io_kwargs=data_io_kwargs
            )
        # But temporarily skipping LabeledEvents
        elif is_ndx_events_installed and isinstance(neurodata_object, ndx_events.LabeledEvents):
            continue
        # Skip the setting of a DataIO when target dataset is a link (assume it will be found in parent)
        elif isinstance(neurodata_object, TimeSeries) and is_dataset_linked:
            continue
        # Strictly speaking, it would be odd if a `backend_configuration` got to this line, but might as well be safe
        else:
            raise NotImplementedError(
                f"Unsupported object type {type(neurodata_object)} for backend configuration "
                f"of {neurodata_object.name}!"
            )



"""
Converters are easy-to-use all-in-one classes for handling multi-stream data.

They are collections of particular data interfaces that commonly occur together (such as AP/LF bands of SpikeGLX
or one photon and behavior videos of Miniscope), yet the objects behave just as a normal interfaces do (multiple
converters can still be combined into a parent converter for a dataset).
"""

from ..datainterfaces.behavior.lightningpose.lightningposeconverter import (
    LightningPoseConverter,
)
from ..datainterfaces.ecephys.spikeglx.spikeglxconverter import SpikeGLXConverterPipe
from ..datainterfaces.ophys.brukertiff.brukertiffconverter import (
    BrukerTiffMultiPlaneConverter,
    BrukerTiffSinglePlaneConverter,
)
from ..datainterfaces.ophys.miniscope.miniscopeconverter import MiniscopeConverter

converter_list = [
    LightningPoseConverter,
    SpikeGLXConverterPipe,
    BrukerTiffMultiPlaneConverter,
    BrukerTiffSinglePlaneConverter,
    MiniscopeConverter,
]



"""DataInterfaces for SpikeGLX."""

from pathlib import Path
from typing import Optional

import numpy as np
from pydantic import FilePath, validate_call

from .spikeglx_utils import (
    add_recording_extractor_properties,
    fetch_stream_id_for_spikelgx_file,
    get_device_metadata,
    get_session_start_time,
)
from ..baserecordingextractorinterface import BaseRecordingExtractorInterface
from ....utils import get_schema_from_method_signature


class SpikeGLXRecordingInterface(BaseRecordingExtractorInterface):
    """
    Primary SpikeGLX interface for converting raw SpikeGLX data using a :py:class:`~spikeinterface.extractors.SpikeGLXRecordingExtractor`.
    """

    display_name = "SpikeGLX Recording"
    keywords = BaseRecordingExtractorInterface.keywords + ("Neuropixels",)
    associated_suffixes = (".imec{probe_index}", ".ap", ".lf", ".meta", ".bin")
    info = "Interface for SpikeGLX recording data."

    # TODO: Add probe_index to probeinterface and propagate it from there
    # Note to developer.
    # In a conversion with Jennifer Colonell she refers to the number after imec as the probe index
    # Quoting here:
    # imec0 is the probe in the lowest slot and port number, imec1 in the next highest, and so on.
    # If you have probes in {slot 2, port 3}, {slot 3, port1} and {slot3, port2},
    # the probe indices in the SGLX output will be 0, 1, and 2, respectively.

    ExtractorName = "SpikeGLXRecordingExtractor"

    @classmethod
    def get_source_schema(cls) -> dict:
        source_schema = get_schema_from_method_signature(method=cls.__init__, exclude=["x_pitch", "y_pitch"])
        source_schema["properties"]["file_path"]["description"] = "Path to SpikeGLX ap.bin or lf.bin file."
        return source_schema

    @validate_call
    def __init__(
        self,
        file_path: FilePath,
        verbose: bool = True,
        es_key: Optional[str] = None,
    ):
        """
        Parameters
        ----------
        file_path : FilePathType
            Path to .bin file. Point to .ap.bin for SpikeGLXRecordingInterface and .lf.bin for SpikeGLXLFPInterface.
        verbose : bool, default: True
            Whether to output verbose text.
        es_key : str, default: "ElectricalSeries"
        """

        self.stream_id = fetch_stream_id_for_spikelgx_file(file_path)
        if es_key is None:
            if "lf" in self.stream_id:
                es_key = "ElectricalSeriesLF"
            elif "ap" in self.stream_id:
                es_key = "ElectricalSeriesAP"
            else:
                raise ValueError("Cannot automatically determine es_key from path")
        file_path = Path(file_path)
        folder_path = file_path.parent
        super().__init__(
            folder_path=folder_path,
            stream_id=self.stream_id,
            verbose=verbose,
            es_key=es_key,
            all_annotations=True,
        )
        self.source_data["file_path"] = str(file_path)
        self.meta = self.recording_extractor.neo_reader.signals_info_dict[(0, self.stream_id)]["meta"]

        # Set electrodes properties
        add_recording_extractor_properties(self.recording_extractor)

    def get_metadata(self) -> dict:
        metadata = super().get_metadata()
        session_start_time = get_session_start_time(self.meta)
        if session_start_time:
            metadata["NWBFile"]["session_start_time"] = session_start_time

        # Device metadata
        device = get_device_metadata(self.meta)

        # Should follow pattern 'Imec0', 'Imec1', etc.
        probe_name = self.stream_id[:5].capitalize()
        device["name"] = f"Neuropixel{probe_name}"

        # Add groups metadata
        metadata["Ecephys"]["Device"] = [device]
        electrode_groups = [
            dict(
                name=group_name,
                description=f"A group representing probe/shank '{group_name}'.",
                location="unknown",
                device=device["name"],
            )
            for group_name in set(self.recording_extractor.get_property("group_name"))
        ]
        metadata["Ecephys"]["ElectrodeGroup"] = electrode_groups

        # Electrodes columns descriptions
        metadata["Ecephys"]["Electrodes"] = [
            dict(name="group_name", description="Name of the ElectrodeGroup this electrode is a part of."),
            dict(name="contact_shapes", description="The shape of the electrode"),
            dict(name="contact_ids", description="The id of the contact on the electrode"),
        ]

        if self.recording_extractor.get_probe().get_shank_count() > 1:
            metadata["Ecephys"]["Electrodes"].append(
                dict(name="shank_ids", description="The shank id of the electrode")
            )

        return metadata

    def get_original_timestamps(self) -> np.ndarray:
        new_recording = self.get_extractor()(
            folder_path=self.source_data["folder_path"], stream_id=self.source_data["stream_id"]
        )  # TODO: add generic method for aliasing from NeuroConv signature to SI init
        if self._number_of_segments == 1:
            return new_recording.get_times()
        else:
            return [
                new_recording.get_times(segment_index=segment_index)
                for segment_index in range(self._number_of_segments)
            ]



"""Collection of all helper functions that require at least one external dependency (some being optional as well)."""

from .importing import (
    get_format_summaries,
    get_package,
    get_package_version,
    is_package_installed,
)
from .nwb_helpers import get_module
from .path_expansion import LocalPathExpander
from .processes import deploy_process



from pydantic import FilePath

from ..baserecordingextractorinterface import BaseRecordingExtractorInterface


class BiocamRecordingInterface(BaseRecordingExtractorInterface):
    """
    Primary data interface class for converting Biocam data.

    Using the :py:class:`~spikeinterface.extractors.BiocamRecordingExtractor`.
    """

    display_name = "Biocam Recording"
    associated_suffixes = (".bwr",)
    info = "Interface for Biocam recording data."

    @classmethod
    def get_source_schema(cls) -> dict:
        schema = super().get_source_schema()
        schema["properties"]["file_path"]["description"] = "Path to the .bwr file."
        return schema

    def __init__(self, file_path: FilePath, verbose: bool = True, es_key: str = "ElectricalSeries"):
        """
        Load and prepare data for Biocam.

        Parameters
        ----------
        file_path : string or Path
            Path to the .bwr file.
        verbose : bool, default: True
            Allows verbose.
        es_key: str, default: "ElectricalSeries"
        """
        super().__init__(file_path=file_path, verbose=verbose, es_key=es_key)



import importlib
import pickle
import warnings
from pathlib import Path
from typing import Optional, Union

import numpy as np
import pandas as pd
import yaml
from pydantic import FilePath
from pynwb import NWBFile
from ruamel.yaml import YAML


def _read_config(config_file_path: FilePath) -> dict:
    """
    Reads structured config file defining a project.
    """

    ruamelFile = YAML()
    path = Path(config_file_path)

    if not path.exists():
        raise FileNotFoundError(f"Config file {path} not found.")

    try:
        with open(path, "r") as f:
            cfg = ruamelFile.load(f)
            curr_dir = config_file_path.parent
            if cfg["project_path"] != curr_dir:
                cfg["project_path"] = curr_dir
    except Exception as err:
        if len(err.args) > 2:
            if err.args[2] == "could not determine a constructor for the tag '!!python/tuple'":
                with open(path, "r") as ymlfile:
                    cfg = yaml.load(ymlfile, Loader=yaml.SafeLoader)
            else:
                raise

    return cfg


def _get_cv2_timestamps(file_path: Union[Path, str]):
    """
    Extract and return an array of timestamps for each frame in a video using OpenCV.

    Parameters
    ----------
    file_path : Union[Path, str]
        The path to the video file from which to extract timestamps.

    Returns
    -------
    np.ndarray
        A numpy array containing the timestamps (in milliseconds) for each frame in the video.
    """
    import cv2
    from tqdm.auto import tqdm

    reader = cv2.VideoCapture(file_path)
    timestamps = []
    n_frames = int(reader.get(cv2.CAP_PROP_FRAME_COUNT))
    fps = reader.get(cv2.CAP_PROP_FPS)

    # Calculate total time in minutes
    total_time_minutes = n_frames / (fps * 60)

    description = (
        "Inferring timestamps from video.\n"
        "This step can be avoided by previously setting the timestamps with `set_aligned_timestamps`"
    )
    tqdm.write(description)

    # Use tqdm with a context manager
    with tqdm(
        total=n_frames,
        desc="Processing frames",
        unit=" frame",
        bar_format="{l_bar}{bar} | {rate_fmt} | {elapsed}<{remaining} | {unit_divisor}",
    ) as pbar:
        # Note, using unit_divisor instead of postfix because of a bug in tqdm https://github.com/tqdm/tqdm/issues/712

        for frame_idx in range(n_frames):
            _ = reader.read()
            timestamps.append(reader.get(cv2.CAP_PROP_POS_MSEC))

            current_time_minutes = (frame_idx + 1) / (fps * 60)

            pbar.unit_divisor = f"processed {current_time_minutes:.2f} of {total_time_minutes:.2f} minutes"
            pbar.update(1)

    reader.release()
    return timestamps


def _get_movie_timestamps(movie_file, VARIABILITYBOUND=1000, infer_timestamps=True):
    """
    Return numpy array of the timestamps for a video.

    Parameters
    ----------
    movie_file : str
        Path to movie_file
    """

    timestamps = _get_cv2_timestamps(file_path=movie_file)
    timestamps = np.array(timestamps) / 1000  # Convert to seconds

    import cv2

    reader = cv2.VideoCapture(movie_file)
    fps = reader.get(cv2.CAP_PROP_FPS)
    reader.release()

    if np.nanvar(np.diff(timestamps)) < 1.0 / fps * 1.0 / VARIABILITYBOUND:
        warnings.warn(
            "Variability of timestamps suspiciously small. See: https://github.com/DeepLabCut/DLC2NWB/issues/1"
        )

    if any(timestamps[1:] == 0):
        # Infers times when OpenCV provides 0s
        warning_msg = "Removing"
        timestamp_zero_count = np.count_nonzero(timestamps == 0)
        timestamps[1:][timestamps[1:] == 0] = np.nan  # replace 0s with nan

        if infer_timestamps:
            warning_msg = "Replacing"
            timestamps = _infer_nan_timestamps(timestamps)

        warnings.warn(  # warns user of percent of 0 frames
            "%s cv2 timestamps returned as 0: %f%%" % (warning_msg, (timestamp_zero_count / len(timestamps) * 100))
        )

    return timestamps


def _infer_nan_timestamps(timestamps):
    """Given np.array, interpolate nan values using index * sampling rate"""
    bad_timestamps_mask = np.isnan(timestamps)
    # Runs of good timestamps
    good_run_indices = np.where(np.diff(np.hstack(([False], bad_timestamps_mask == False, [False]))))[0].reshape(-1, 2)

    # For each good run, get the diff and append to cumulative array
    sampling_diffs = np.array([])
    for idx in good_run_indices:
        sampling_diffs = np.append(sampling_diffs, np.diff(timestamps[idx[0] : idx[1]]))
    estimated_sampling_rate = np.mean(sampling_diffs)  # Average over diffs

    # Infer timestamps with avg sampling rate
    bad_timestamps_indexes = np.argwhere(bad_timestamps_mask)[:, 0]
    inferred_timestamps = bad_timestamps_indexes * estimated_sampling_rate
    timestamps[bad_timestamps_mask] = inferred_timestamps

    return timestamps


def _ensure_individuals_in_header(df, individual_name: str):
    """
    Ensure that the 'individuals' column is present in the header of the given DataFrame.

    Parameters:
        df (pandas.DataFrame): The DataFrame to modify.
        individual_name (str): The name of the individual to add to the header.

    Returns:
        pandas.DataFrame: The modified DataFrame with the 'individuals' column added to the header.

    Notes:
        - If the 'individuals' column is already present in the header, no modifications are made.
        - If the 'individuals' column is not present, a new DataFrame is created with the 'individual_name'
        as the column name, and the 'individuals' column is added to the header of the DataFrame.
        - The order of the columns in the header is preserved.

    """
    if "individuals" not in df.columns.names:
        # Single animal project -> add individual row to
        # the header of single animal projects.
        temp = pd.concat({individual_name: df}, names=["individuals"], axis=1)
        df = temp.reorder_levels(["scorer", "individuals", "bodyparts", "coords"], axis=1)

    return df


def _get_graph_edges(metadata_file_path: Path):
    """
    Extracts the part affinity field graph from the metadata pickle file.

    Parameters
    ----------
    metadata_file_path : Path
        The path to the metadata pickle file.

    Returns
    -------
    list
        The part affinity field graph, which defines the edges between the keypoints in the pose estimation.
    """
    paf_graph = []
    if metadata_file_path.exists():
        with open(metadata_file_path, "rb") as file:
            metadata = pickle.load(file)

        test_cfg = metadata["data"]["DLC-model-config file"]
        paf_graph = test_cfg.get("partaffinityfield_graph", [])
        if paf_graph:
            paf_inds = test_cfg.get("paf_best")
            if paf_inds is not None:
                paf_graph = [paf_graph[i] for i in paf_inds]
    else:
        warnings.warn("Metadata not found...")

    return paf_graph


def _get_video_info_from_config_file(config_file_path: Path, vidname: str):
    """
    Get the video information from the project config file.

    Parameters
    ----------
    config_file_path : Path
        The path to the project config file.
    vidname : str
        The name of the video.

    Returns
    -------
    tuple
        A tuple containing the video file path and the image shape.
    """
    config_file_path = Path(config_file_path)
    cfg = _read_config(config_file_path)

    video = None
    for video_path, params in cfg["video_sets"].items():
        if vidname in video_path:
            video = video_path, params["crop"]
            break

    if video is None:
        warnings.warn(f"The corresponding video file could not be found in the config file")
        video = None, "0, 0, 0, 0"

    # The video in the config_file looks like this:
    # video_sets:
    #    /Data/openfield-Pranav-2018-08-20/videos/m1s1.mp4:
    #        crop: 0, 640, 0, 480

    video_file_path, image_shape = video

    return video_file_path, image_shape


def _get_pes_args(
    *,
    h5file: Path,
    individual_name: str,
):
    h5file = Path(h5file)

    _, scorer = h5file.stem.split("DLC")
    scorer = "DLC" + scorer

    df = _ensure_individuals_in_header(pd.read_hdf(h5file), individual_name)

    return scorer, df


def _write_pes_to_nwbfile(
    nwbfile,
    animal,
    df_animal,
    scorer,
    video_file_path,
    image_shape,
    paf_graph,
    timestamps,
    exclude_nans,
    pose_estimation_container_kwargs: Optional[dict] = None,
):

    from ndx_pose import PoseEstimation, PoseEstimationSeries

    pose_estimation_container_kwargs = pose_estimation_container_kwargs or dict()

    pose_estimation_series = []
    for keypoint in df_animal.columns.get_level_values("bodyparts").unique():
        data = df_animal.xs(keypoint, level="bodyparts", axis=1).to_numpy()

        if exclude_nans:
            # exclude_nans is inverse infer_timestamps. if not infer, there may be nans
            data = data[~np.isnan(timestamps)]
            timestamps_cleaned = timestamps[~np.isnan(timestamps)]
        else:
            timestamps_cleaned = timestamps

        pes = PoseEstimationSeries(
            name=f"{animal}_{keypoint}" if animal else keypoint,
            description=f"Keypoint {keypoint} from individual {animal}.",
            data=data[:, :2],
            unit="pixels",
            reference_frame="(0,0) corresponds to the bottom left corner of the video.",
            timestamps=timestamps_cleaned,
            confidence=data[:, 2],
            confidence_definition="Softmax output of the deep neural network.",
        )
        pose_estimation_series.append(pes)

    deeplabcut_version = None
    is_deeplabcut_installed = importlib.util.find_spec(name="deeplabcut") is not None
    if is_deeplabcut_installed:
        deeplabcut_version = importlib.metadata.version(distribution_name="deeplabcut")

    # TODO, taken from the original implementation, improve it if the video is passed
    dimensions = [list(map(int, image_shape.split(",")))[1::2]]
    pose_estimation_default_kwargs = dict(
        pose_estimation_series=pose_estimation_series,
        description="2D keypoint coordinates estimated using DeepLabCut.",
        original_videos=[video_file_path],
        dimensions=dimensions,
        scorer=scorer,
        source_software="DeepLabCut",
        source_software_version=deeplabcut_version,
        nodes=[pes.name for pes in pose_estimation_series],
        edges=paf_graph if paf_graph else None,
        **pose_estimation_container_kwargs,
    )
    pose_estimation_default_kwargs.update(pose_estimation_container_kwargs)
    pose_estimation_container = PoseEstimation(**pose_estimation_default_kwargs)

    if "behavior" in nwbfile.processing:  # TODO: replace with get_module
        behavior_processing_module = nwbfile.processing["behavior"]
    else:
        behavior_processing_module = nwbfile.create_processing_module(
            name="behavior", description="processed behavioral data"
        )
    behavior_processing_module.add(pose_estimation_container)

    return nwbfile


def add_subject_to_nwbfile(
    nwbfile: NWBFile,
    h5file: FilePath,
    individual_name: str,
    config_file: Optional[FilePath] = None,
    timestamps: Optional[Union[list, np.ndarray]] = None,
    pose_estimation_container_kwargs: Optional[dict] = None,
) -> NWBFile:
    """
    Given the subject name, add the DLC .h5 file to an in-memory NWBFile object.

    Parameters
    ----------
    nwbfile : pynwb.NWBFile
        The in-memory nwbfile object to which the subject specific pose estimation series will be added.
    h5file : str or path
        Path to the DeepLabCut .h5 output file.
    individual_name : str
        Name of the subject (whose pose is predicted) for single-animal DLC project.
        For multi-animal projects, the names from the DLC project will be used directly.
    config_file : str or path, optional
        Path to a project config.yaml file
    timestamps : list, np.ndarray or None, default: None
        Alternative timestamps vector. If None, then use the inferred timestamps from DLC2NWB
    pose_estimation_container_kwargs : dict, optional
        Dictionary of keyword argument pairs to pass to the PoseEstimation container.

    Returns
    -------
    nwbfile : pynwb.NWBFile
        nwbfile with pes written in the behavior module
    """
    h5file = Path(h5file)

    if "DLC" not in h5file.name or not h5file.suffix == ".h5":
        raise IOError("The file passed in is not a DeepLabCut h5 data file.")

    video_name, scorer = h5file.stem.split("DLC")
    scorer = "DLC" + scorer

    # TODO probably could be read directly with h5py
    # This requires pytables
    data_frame_from_hdf5 = pd.read_hdf(h5file)
    df = _ensure_individuals_in_header(data_frame_from_hdf5, individual_name)

    # Note the video here is a tuple of the video path and the image shape
    if config_file is not None:
        video_file_path, image_shape = _get_video_info_from_config_file(
            config_file_path=config_file,
            vidname=video_name,
        )
    else:
        video_file_path = None
        image_shape = "0, 0, 0, 0"

    # find timestamps only if required:``
    timestamps_available = timestamps is not None
    if not timestamps_available:
        if video_file_path is None:
            timestamps = df.index.tolist()  # setting timestamps to dummy
        else:
            timestamps = _get_movie_timestamps(video_file_path, infer_timestamps=True)

    # Fetch the corresponding metadata pickle file, we extract the edges graph from here
    # TODO: This is the original implementation way to extract the file name but looks very brittle. Improve it
    filename = str(h5file.parent / h5file.stem)
    for i, c in enumerate(filename[::-1]):
        if c.isnumeric():
            break
    if i > 0:
        filename = filename[:-i]

    metadata_file_path = Path(filename + "_meta.pickle")
    paf_graph = _get_graph_edges(metadata_file_path=metadata_file_path)

    df_animal = df.xs(individual_name, level="individuals", axis=1)

    return _write_pes_to_nwbfile(
        nwbfile,
        individual_name,
        df_animal,
        scorer,
        video_file_path,
        image_shape,
        paf_graph,
        timestamps,
        exclude_nans=False,
        pose_estimation_container_kwargs=pose_estimation_container_kwargs,
    )



from typing import Optional

import pandas as pd
from pydantic import FilePath, validate_call

from ..timeintervalsinterface import TimeIntervalsInterface


class ExcelTimeIntervalsInterface(TimeIntervalsInterface):
    """Interface for adding data from an Excel file to NWB as a TimeIntervals object."""

    display_name = "Excel time interval table"
    associated_suffixes = (".xlsx", ".xls", ".xlsm")
    info = "Interface for writing a time intervals table from an excel file."

    @validate_call
    def __init__(
        self,
        file_path: FilePath,
        read_kwargs: Optional[dict] = None,
        verbose: bool = True,
    ):
        """
        Parameters
        ----------
        file_path : FilePath
        read_kwargs : dict, optional
            Passed to pandas.read_excel()
        verbose : bool, default: True
        """
        super().__init__(file_path=file_path, read_kwargs=read_kwargs, verbose=verbose)

    def _read_file(self, file_path: FilePath, **read_kwargs):
        return pd.read_excel(file_path, **read_kwargs)



import math
from pathlib import Path
from typing import Optional, Union

import numpy as np
from numpy.typing import DTypeLike
from pydantic import DirectoryPath
from pynwb import NWBHDF5IO, H5DataIO, TimeSeries
from pynwb.testing.mock.file import mock_NWBFile

from ..importing import is_package_installed
from ...utils import ArrayType


def _check_parameter_dtype_consistency(
    parameter_name: str,
    parameter_value: Union[int, float],
    generic_dtype: type,  # Literal[np.integer, np.floating]
):
    """Helper for `generate_mock_ttl_signal` to assert consistency between parameters and expected trace dtype."""
    end_format = "an integer" if generic_dtype == np.integer else "a float"
    assert np.issubdtype(type(parameter_value), generic_dtype), (
        f"If specifying the '{parameter_name}' manually, please ensure it matches the 'dtype'! "
        f"Received '{type(parameter_value).__name__}', should be {end_format}."
    )


def generate_mock_ttl_signal(
    signal_duration: float = 7.0,
    ttl_times: Optional[ArrayType] = None,
    ttl_duration: float = 1.0,
    sampling_frequency_hz: float = 25_000.0,
    dtype: DTypeLike = "int16",
    baseline_mean: Optional[Union[int, float]] = None,
    signal_mean: Optional[Union[int, float]] = None,
    channel_noise: Optional[Union[int, float]] = None,
    random_seed: Optional[int] = 0,
) -> np.ndarray:
    """
    Generate a synthetic signal of TTL pulses similar to those seen in .nidq.bin files using SpikeGLX.

    Parameters
    ----------
    signal_duration : float, default: 7.0
        The number of seconds to simulate.
    ttl_times : array of floats, optional
        The times within the `signal_duration` to trigger the TTL pulse.
        In conjunction with the `ttl_duration`, these must produce disjoint 'on' intervals.
        The default generates a periodic 1 second on, 1 second off pattern.
    ttl_duration : float, default: 1.0
        How long the TTL pulse stays in the 'on' state when triggered, in seconds.
        In conjunction with the `ttl_times`, these must produce disjoint 'on' intervals.
    sampling_frequency_hz : float, default: 25,000.0
        The sampling frequency of the signal in Hz.
        The default is 25000 Hz; similar to that of typical .nidq.bin files.
    dtype : numpy data type or one of its accepted string input, default: "int16"
        The data type of the trace.
        Must match the data type of `baseline_mean`, `signal_mean`, and `channel_noise`, if any of those are specified.
        Recommended to be int16 for maximum efficiency, but can also be any size float to represent voltage scalings.
    baseline_mean : integer or float, depending on specified 'dtype', optional
        The average value for the baseline; usually around 0 Volts.
        The default is approximately 0.005645752 Volts, estimated from a real example of a TTL pulse in a .nidq.bin file.
    signal_mean : integer or float, optional
        Type depends on specified 'dtype'. The average value for the signal; usually around 5 Volts.
        The default is approximately 4.980773925 Volts, estimated from a real example of a TTL pulse in a .nidq.bin file.
    channel_noise : integer or float, optional
        Type depends on specified 'dtype'. The standard deviation of white noise in the channel.
        The default is approximately 0.002288818 Volts, estimated from a real example of a TTL pulse in a .nidq.bin file.
    random_seed : int or None, default: 0
        The seed to set for the numpy random number generator.
        Set to None to choose the seed randomly.
        The default is kept at 0 for generating reproducible outputs.

    Returns
    -------
    trace: numpy.ndarray
        The synethic trace representing a channel with TTL pulses.
    """
    dtype = np.dtype(dtype)

    # Default values estimated from real files
    baseline_mean_int16_default = 37
    signal_mean_int16_default = 32642
    channel_noise_int16_default = 15
    default_gain_to_volts = 152.58789062 * 1e-6

    if np.issubdtype(dtype, np.unsignedinteger):
        # If data type is an unsigned integer, increment the signed default values by the midpoint of the unsigned range
        shift = math.floor(np.iinfo(dtype).max / 2)
        baseline_mean_int16_default += shift
        signal_mean_int16_default += shift

    if np.issubdtype(dtype, np.integer):
        baseline_mean = baseline_mean or baseline_mean_int16_default
        signal_mean = signal_mean or signal_mean_int16_default
        channel_noise = channel_noise or channel_noise_int16_default
        generic_dtype = np.integer
    else:
        baseline_mean = baseline_mean or baseline_mean_int16_default * default_gain_to_volts
        signal_mean = signal_mean or signal_mean_int16_default * default_gain_to_volts
        channel_noise = channel_noise or channel_noise_int16_default * default_gain_to_volts
        generic_dtype = np.floating
    parameters_to_check = dict(baseline_mean=baseline_mean, signal_mean=signal_mean, channel_noise=channel_noise)
    for parameter_name, parameter_value in parameters_to_check.items():
        _check_parameter_dtype_consistency(
            parameter_name=parameter_name, parameter_value=parameter_value, generic_dtype=generic_dtype
        )

    np.random.seed(seed=random_seed)
    num_frames = np.ceil(signal_duration * sampling_frequency_hz).astype(int)
    trace = (np.random.randn(num_frames) * channel_noise + baseline_mean).astype(dtype)

    if ttl_times is not None:
        ttl_times = np.array(ttl_times)
    else:
        ttl_times = np.arange(start=1.0, stop=signal_duration, step=2.0)

    assert len(ttl_times) == 1 or not any(  # np.diff errors out when len(ttl_times) < 2
        np.diff(ttl_times) <= ttl_duration
    ), "There are overlapping TTL 'on' intervals! Please specify disjoint on/off periods."

    ttl_start_frames = np.round(ttl_times * sampling_frequency_hz).astype(int)
    num_frames_ttl_duration = np.round(ttl_duration * sampling_frequency_hz).astype(int)
    ttl_intervals = (slice(start, start + num_frames_ttl_duration) for start in ttl_start_frames)

    for ttl_interval in ttl_intervals:
        trace[ttl_interval] += signal_mean

    return trace


def regenerate_test_cases(folder_path: DirectoryPath, regenerate_reference_images: bool = False):  # pragma: no cover
    """
    Regenerate the test cases of the file included in the main testing suite, which is frozen between breaking changes.

    Parameters
    ----------
    folder_path : PathType
        Folder to save the resulting NWB file in. For use in the testing suite, this must be the
        '/test_testing/test_mock_ttl/' subfolder adjacent to the 'test_mock_tt.py' file.
    regenerate_reference_images : bool
        If true, uses the kaleido package with plotly (you may need to install both) to regenerate the images used
        as references in the documentation.
    """
    folder_path = Path(folder_path)

    if regenerate_reference_images:
        assert is_package_installed("plotly") and is_package_installed("kaleido"), (
            "To regenerate the reference images, " "you must install both plotly and kaleido!"
        )
        import plotly.graph_objects as go
        from plotly.subplots import make_subplots

        image_file_path = folder_path / "example_ttl_reference.png"

    nwbfile_path = folder_path / "mock_ttl_examples.nwb"
    compression_options = dict(compression="gzip", compression_opts=9)
    unit = "Volts"
    rate = 1000.0  # For non-default series to produce less data

    nwbfile = mock_NWBFile()

    # Test Case 1: Default
    default_ttl_signal = generate_mock_ttl_signal()
    nwbfile.add_acquisition(
        TimeSeries(
            name="DefaultTTLSignal",
            unit=unit,
            rate=25000.0,
            data=H5DataIO(data=default_ttl_signal, chunks=default_ttl_signal.shape, **compression_options),
        )
    )

    non_default_series = dict()

    # Test Case 2: Irregular short pulses
    irregular_short_pulses = generate_mock_ttl_signal(
        signal_duration=2.5, ttl_times=[0.22, 1.37], ttl_duration=0.25, sampling_frequency_hz=rate
    )
    non_default_series.update(IrregularShortPulses=irregular_short_pulses)

    # Test Case 3: Non-default regular
    non_default_regular = generate_mock_ttl_signal(
        signal_duration=2.7,
        ttl_times=[0.2, 1.2, 2.2],
        ttl_duration=0.3,
        sampling_frequency_hz=rate,
    )
    non_default_series.update(NonDefaultRegular=non_default_regular)

    # Test Case 4: Non-default regular with adjusted means
    non_default_regular_adjusted_means = generate_mock_ttl_signal(
        signal_duration=2.7,
        ttl_times=[0.2, 1.2, 2.2],
        ttl_duration=0.3,
        sampling_frequency_hz=rate,
        baseline_mean=300,
        signal_mean=20000,
    )
    non_default_series.update(NonDefaultRegularAdjustedMeans=non_default_regular_adjusted_means)

    # Test Case 5: Irregular short pulses with adjusted noise
    irregular_short_pulses_adjusted_noise = generate_mock_ttl_signal(
        signal_duration=2.5,
        ttl_times=[0.22, 1.37],
        ttl_duration=0.25,
        sampling_frequency_hz=rate,
        channel_noise=2,
    )
    non_default_series.update(IrregularShortPulsesAdjustedNoise=irregular_short_pulses_adjusted_noise)

    # Test Case 6: Non-default regular as floats
    non_default_regular_as_floats = generate_mock_ttl_signal(
        signal_duration=2.7,
        ttl_times=[0.2, 1.2, 2.2],
        ttl_duration=0.3,
        sampling_frequency_hz=rate,
        dtype="float32",
    )
    non_default_series.update(NonDefaultRegularFloats=non_default_regular_as_floats)

    # Test Case 7: Non-default regular as floats with adjusted means and noise (which are also, then, floats)
    non_default_regular_as_floats_adjusted_means_and_noise = generate_mock_ttl_signal(
        signal_duration=2.7,
        ttl_times=[0.2, 1.2, 2.2],
        ttl_duration=0.3,
        sampling_frequency_hz=rate,
        dtype="float32",
        baseline_mean=1.1,
        signal_mean=7.2,
        channel_noise=0.4,
    )
    non_default_series.update(FloatsAdjustedMeansAndNoise=non_default_regular_as_floats_adjusted_means_and_noise)

    # Test Case 8: Non-default regular as uint16
    non_default_regular_as_uint16 = generate_mock_ttl_signal(
        signal_duration=2.7,
        ttl_times=[0.2, 1.2, 2.2],
        ttl_duration=0.3,
        sampling_frequency_hz=rate,
        dtype="uint16",
    )
    non_default_series.update(NonDefaultRegularUInt16=non_default_regular_as_uint16)

    # Test Case 9: Irregular short pulses with different seed
    irregular_short_pulses_different_seed = generate_mock_ttl_signal(
        signal_duration=2.5,
        ttl_times=[0.22, 1.37],
        ttl_duration=0.25,
        sampling_frequency_hz=rate,
        random_seed=1,
    )
    non_default_series.update(IrregularShortPulsesDifferentSeed=irregular_short_pulses_different_seed)

    if regenerate_reference_images:
        num_cols = 5
        plot_index = 1
        subplot_titles = ["Default"]
        subplot_titles.extend(list(non_default_series))
        fig = make_subplots(rows=2, cols=num_cols, subplot_titles=subplot_titles)
        fig.add_trace(go.Scatter(y=default_ttl_signal, text="Default"), row=1, col=1)

    for time_series_name, time_series_data in non_default_series.items():
        nwbfile.add_acquisition(
            TimeSeries(
                name=time_series_name,
                unit=unit,
                rate=rate,
                data=H5DataIO(data=time_series_data, chunks=time_series_data.shape, **compression_options),
            )
        )

        if regenerate_reference_images:
            fig.add_trace(
                go.Scatter(y=time_series_data, text=time_series_name),
                row=math.floor(plot_index / num_cols) + 1,
                col=int(plot_index % num_cols) + 1,
            )
            plot_index += 1

    if regenerate_reference_images:
        fig.update_annotations(font_size=6)
        fig.update_layout(showlegend=False)
        fig.update_yaxes(tickfont=dict(size=5))
        fig.update_xaxes(showticklabels=False)
        fig.write_image(file=image_file_path)

    with NWBHDF5IO(path=nwbfile_path, mode="w") as io:
        io.write(nwbfile)



import platform
import sys
from collections import defaultdict
from pathlib import Path
from shutil import copy

from setuptools import setup

root = Path(__file__).parent


def read_requirements(file):
    """Read requirements from a file."""
    with open(root / file) as f:
        return f.readlines()


extras_require = defaultdict(list)
extras_require["full"] = ["dandi>=0.58.1", "hdf5plugin", "boto3"]

for modality in ["ophys", "ecephys", "icephys", "behavior", "text"]:
    modality_path = root / "src" / "neuroconv" / "datainterfaces" / modality
    modality_requirement_file = modality_path / "requirements.txt"
    if modality_requirement_file.exists():
        modality_requirements = read_requirements(modality_requirement_file)
        extras_require["full"].extend(modality_requirements)
        extras_require[modality] = modality_requirements
    else:
        modality_requirements = []

    format_subpaths = [path for path in modality_path.iterdir() if path.is_dir() and path.name != "__pycache__"]
    for format_subpath in format_subpaths:
        format_requirement_file = format_subpath / "requirements.txt"
        extras_require[format_subpath.name] = modality_requirements.copy()
        if format_requirement_file.exists():
            format_requirements = read_requirements(format_requirement_file)
            extras_require["full"].extend(format_requirements)
            extras_require[modality].extend(format_requirements)
            extras_require[format_subpath.name].extend(format_requirements)

# Create a local copy for the gin test configuration file based on the master file `base_gin_test_config.json`
gin_config_file_base = root / "base_gin_test_config.json"
gin_config_file_local = root / "tests/test_on_data/gin_test_config.json"
if not gin_config_file_local.exists():
    gin_config_file_local.parent.mkdir(parents=True, exist_ok=True)
    copy(src=gin_config_file_base, dst=gin_config_file_local)

# Bug related to sonpy on M1 Mac being installed but not running properly
if sys.platform == "darwin" and platform.processor() == "arm":
    extras_require.pop("spike2", None)
    extras_require["ecephys"] = [req for req in extras_require["ecephys"] if "sonpy" not in req]
    extras_require["full"] = [req for req in extras_require["full"] if "sonpy" not in req]

setup(
    extras_require=extras_require,
)



from pathlib import Path
from typing import Optional

from pydantic import DirectoryPath, validate_call

from .spikeglxdatainterface import SpikeGLXRecordingInterface
from .spikeglxnidqinterface import SpikeGLXNIDQInterface
from ....nwbconverter import ConverterPipe
from ....utils import get_schema_from_method_signature


class SpikeGLXConverterPipe(ConverterPipe):
    """
    The simplest, easiest to use class for converting all SpikeGLX data in a folder.

    Primary conversion class for handling multiple SpikeGLX data streams.
    """

    display_name = "SpikeGLX Converter"
    keywords = SpikeGLXRecordingInterface.keywords + SpikeGLXNIDQInterface.keywords
    associated_suffixes = SpikeGLXRecordingInterface.associated_suffixes + SpikeGLXNIDQInterface.associated_suffixes
    info = "Converter for multi-stream SpikeGLX recording data."

    @classmethod
    def get_source_schema(cls):
        source_schema = get_schema_from_method_signature(method=cls.__init__, exclude=["streams"])
        source_schema["properties"]["folder_path"]["description"] = "Path to the folder containing SpikeGLX streams."
        return source_schema

    @classmethod
    def get_streams(cls, folder_path: DirectoryPath) -> list[str]:
        from spikeinterface.extractors import SpikeGLXRecordingExtractor

        return SpikeGLXRecordingExtractor.get_streams(folder_path=folder_path)[0]

    @validate_call
    def __init__(
        self,
        folder_path: DirectoryPath,
        streams: Optional[list[str]] = None,
        verbose: bool = False,
    ):
        """
        Read all data from multiple streams stored in the SpikeGLX format.

        This can include...
            (a) single-probe but multi-band such as AP+LF streams
            (b) multi-probe with multi-band
            (c) with or without the associated NIDQ channels

        Parameters
        ----------
        folder_path : DirectoryPath
            Path to folder containing the NIDQ stream and subfolders containing each IMEC stream.
        streams : list of strings, optional
            A specific list of streams you wish to load.
            To see which streams are available, run `SpikeGLXConverter.get_streams(folder_path="path/to/spikeglx")`.
            By default, all available streams are loaded.
        verbose : bool, default: False
            Whether to output verbose text.
        """
        folder_path = Path(folder_path)

        streams = streams or self.get_streams(folder_path=folder_path)

        data_interfaces = dict()
        for stream in streams:
            if "ap" in stream:
                probe_name = stream[:5]
                file_path = (
                    folder_path / f"{folder_path.stem}_{probe_name}" / f"{folder_path.stem}_t0.{probe_name}.ap.bin"
                )
                es_key = f"ElectricalSeriesAP{probe_name.capitalize()}"
                interface = SpikeGLXRecordingInterface(file_path=file_path, es_key=es_key)
            elif "lf" in stream:
                probe_name = stream[:5]
                file_path = (
                    folder_path / f"{folder_path.stem}_{probe_name}" / f"{folder_path.stem}_t0.{probe_name}.lf.bin"
                )
                es_key = f"ElectricalSeriesLF{probe_name.capitalize()}"
                interface = SpikeGLXRecordingInterface(file_path=file_path, es_key=es_key)
            elif "nidq" in stream:
                file_path = folder_path / f"{folder_path.stem}_t0.nidq.bin"
                interface = SpikeGLXNIDQInterface(file_path=file_path)
            data_interfaces.update({str(stream): interface})  # Without str() casting, is a numpy string

        super().__init__(data_interfaces=data_interfaces, verbose=verbose)

    def get_conversion_options_schema(self) -> dict:
        conversion_options_schema = super().get_conversion_options_schema()
        conversion_options_schema["properties"].update(
            {name: interface.get_conversion_options_schema() for name, interface in self.data_interface_objects.items()}
        )
        return conversion_options_schema



from pathlib import Path

from pydantic import DirectoryPath, validate_call
from pynwb import NWBFile

from .... import BaseDataInterface
from ....tools import get_package
from ....utils import DeepDict


class MiniscopeBehaviorInterface(BaseDataInterface):
    """Data Interface for Miniscope behavior data."""

    display_name = "Miniscope Behavior"
    keywords = ("video",)
    associated_suffixes = (".avi",)
    info = "Interface for Miniscope behavior video data."

    @classmethod
    def get_source_schema(cls) -> dict:
        source_schema = super().get_source_schema()
        source_schema["properties"]["folder_path"][
            "description"
        ] = "The main Miniscope folder. The movie files are expected to be in sub folders within the main folder."
        return source_schema

    @validate_call
    def __init__(self, folder_path: DirectoryPath):
        """
        Initialize reading recordings from the Miniscope behavioral camera.

        Parameters
        ----------
        folder_path : DirectoryPath
            The path that points to the main Miniscope folder.
            The movie files are expected to be in sub folders within the main folder.
        """
        from ndx_miniscope.utils import (
            get_recording_start_times,
            get_starting_frames,
            get_timestamps,
            read_miniscope_config,
        )

        natsort = get_package(package_name="natsort", installation_instructions="pip install natsort")

        super().__init__(folder_path=folder_path)

        folder_path = Path(self.source_data["folder_path"])
        self._behav_avi_file_paths = natsort.natsorted(list(folder_path.glob("*/BehavCam*/*.avi")))
        assert (
            self._behav_avi_file_paths
        ), f"The behavior movies (.avi files) are missing from '{self.source_data['folder_path']}'."
        # note we might have to change this if we have other examples
        configuration_file_name = "metaData.json"
        miniscope_config_files = natsort.natsorted(list(folder_path.glob(f"*/BehavCam*/{configuration_file_name}")))
        assert (
            miniscope_config_files
        ), f"The configuration files ({configuration_file_name} files) are missing from '{folder_path}'."

        behavcam_subfolders = list(folder_path.glob("*/BehavCam*/"))
        self._miniscope_config = read_miniscope_config(folder_path=str(behavcam_subfolders[0]))

        self._recording_start_times = get_recording_start_times(folder_path=str(folder_path))
        self._starting_frames = get_starting_frames(folder_path=str(folder_path))
        assert len(self._starting_frames) == len(self._behav_avi_file_paths)
        self._timestamps = get_timestamps(folder_path=str(folder_path), file_pattern="BehavCam*/timeStamps.csv")

    def get_metadata(self) -> DeepDict:
        metadata = super().get_metadata()
        metadata["NWBFile"].update(session_start_time=self._recording_start_times[0])

        metadata["Behavior"]["Device"] = [self._miniscope_config]
        # Add link to Device for ImageSeries
        metadata["Behavior"]["ImageSeries"] = [
            dict(
                name="BehavCamImageSeries",
                device=self._miniscope_config["name"],
                dimension=[self._miniscope_config["ROI"]["width"], self._miniscope_config["ROI"]["height"]],
                unit="px",
            )
        ]

        return metadata

    def add_to_nwbfile(
        self,
        nwbfile: NWBFile,
        metadata: DeepDict,
    ):
        """
        Adds the behavior movies as ImageSeries from provided metadata.
        The created ImageSeries is linked to a Miniscope device.

        Parameters
        ----------
        nwbfile : NWBFile
            The nwbfile to add the image series to.
        metadata : DeepDict
            The metadata storing the necessary metadata for creating the image series and linking it to the appropriate device.
            The metadata for the image series should be stored in metadata["Behavior"]["ImageSeries"].
            The metadata for the device to be linked should be stored in metadata["Behavior"]["Device"].
        """
        from ndx_miniscope.utils import add_miniscope_image_series

        add_miniscope_image_series(
            nwbfile=nwbfile,
            metadata=metadata,
            external_files=[str(file_path) for file_path in self._behav_avi_file_paths],
            starting_frames=self._starting_frames,
            timestamps=self._timestamps,
        )



from typing import Literal

from pydantic import FilePath, validate_call

from ..baseimagingextractorinterface import BaseImagingExtractorInterface


class SbxImagingInterface(BaseImagingExtractorInterface):
    """Data Interface for SbxImagingExtractor."""

    display_name = "Scanbox Imaging"
    associated_suffixes = (".sbx",)
    info = "Interface for Scanbox imaging data."

    @validate_call
    def __init__(
        self,
        file_path: FilePath,
        sampling_frequency: float = None,
        verbose: bool = True,
        photon_series_type: Literal["OnePhotonSeries", "TwoPhotonSeries"] = "TwoPhotonSeries",
    ):
        """
        Parameters
        ----------
        file_path : FilePathType
            Path to .sbx file.
        sampling_frequency : float, optional
        verbose : bool, default: True
        """

        super().__init__(
            file_path=file_path,
            sampling_frequency=sampling_frequency,
            verbose=verbose,
            photon_series_type=photon_series_type,
        )

    def get_metadata(self) -> dict:
        metadata = super().get_metadata()
        metadata["Ophys"]["Device"][0]["description"] = "Scanbox imaging"
        return metadata



"""Collection of helper functions for assessing and performing automated data transfers."""

from ._dandi import automatic_dandi_upload
from ._globus import get_globus_dataset_content_sizes, transfer_globus_content

__all__ = [
    "automatic_dandi_upload",
    "get_globus_dataset_content_sizes",
    "transfer_globus_content",
]



"""Helpful classes for expanding file or folder paths on a system given an f-string rule for matching patterns."""

import abc
import os
from datetime import date, datetime
from pathlib import Path
from typing import Iterable

from parse import parse
from pydantic import DirectoryPath, FilePath

from ..utils import DeepDict


class AbstractPathExpander(abc.ABC):
    """
    Abstract base class for expanding file paths and extracting metadata.

    This class provides methods to extract metadata from file paths within a directory
    and to expand paths based on a specified data specification. It is designed to be
    subclassed, with the `list_directory` method needing to be implemented by any
    subclass to provide the specific logic for listing files in a directory.
    """

    def extract_metadata(self, base_directory: DirectoryPath, format_: str):
        """
        Uses the parse library to extract metadata from file paths in the base_directory.

        This method iterates over files in `base_directory`, parsing each file path according to `format_`.
        The format string is adjusted to the current operating system's path separator. The method yields
        each file path and its corresponding parsed metadata. To constrain metadata matches to only the
        name of the file or folder/directory, the method checks that the metadata does not contain the
        OS path separator (e.g., '/' or '\\').

        Parameters
        ----------
        base_directory : DirectoryPath
            The base directory from which to list files for metadata extraction. It should be a path-like
            object that is convertible to a `pathlib.Path`.
        format_ : str
            The format string used for parsing the file paths. This string can represent a path in any
            OS format, and is adjusted internally to match the current OS's path separator.

        Yields
        ------
        tuple[Path, dict[str, Any]]
            A tuple containing the file path as a `Path` object and a dictionary of the named metadata
            extracted from the file path.
        """

        format_ = format_.replace("\\", os.sep)  # Actual character is a single back-slash; first is an escape for that
        format_ = format_.replace("/", os.sep)  # our f-string uses '/' to communicate os-independent separators

        for filepath in self.list_directory(base_directory=Path(base_directory)):
            result = parse(format_, filepath)
            if result:
                named_result = result.named
                no_field_in_metadata_contains_os_sep = all(os.sep not in str(val) for val in named_result.values())
                if no_field_in_metadata_contains_os_sep:
                    yield filepath, named_result

    @abc.abstractmethod
    def list_directory(self, base_directory: DirectoryPath) -> Iterable[FilePath]:
        """
        List all folders and files in a directory recursively.

        Parameters
        ----------
        base_directory : DirectoryPath
            The base directory whose contents will be iterated recursively.

        Yields
        ------
        sub_paths : iterable of strings
            Generator that yields all sub-paths of file and folders from the common root `base_directory`.
        """
        pass

    def expand_paths(self, source_data_spec: dict[str, dict]) -> list[DeepDict]:
        """
        Match paths in a directory to specs and extract metadata from the paths.

        Parameters
        ----------
        source_data_spec : dict
            Source spec.

        Returns
        -------
        deep_dicts : list of DeepDict objects

        Examples
        --------
        >>> path_expander.expand_paths(
        ...     dict(
        ...         spikeglx=dict(
        ...             base_directory="source_folder",
        ...             paths=dict(
        ...                 file_path="sub-{subject_id}/sub-{subject_id}_ses-{session_id}"
        ...             )
        ...         )
        ...     )
        ... )
        """

        non_standard_super = "extras"
        standard_metadata = {"session_id": "NWBFile", "session_start_time": "NWBFile", "subject_id": "Subject"}

        out = DeepDict()
        for interface, source_data in source_data_spec.items():
            base_directory = Path(source_data["base_directory"]).resolve()
            for path_type in ("file_path", "folder_path"):
                if path_type not in source_data:
                    continue

                _format = source_data[path_type]
                extracted_metadata = self.extract_metadata(base_directory, _format)
                for path, metadata in extracted_metadata:
                    key = tuple((k, v) for k, v in sorted(metadata.items()))

                    asset_path = base_directory / path

                    if path_type == "file_path" and not asset_path.is_file():
                        continue
                    if path_type == "folder_path" and not asset_path.is_dir():
                        continue

                    out[key]["source_data"][interface][path_type] = str(asset_path)

                    for meta_key, meta_val in metadata.items():
                        super_key = standard_metadata.get(meta_key, non_standard_super)
                        if meta_key == "session_start_time" and isinstance(meta_val, date):
                            meta_val = datetime(meta_val.year, meta_val.month, meta_val.day)
                        out[key]["metadata"][super_key][meta_key] = meta_val

        return list(dict(out).values())


class LocalPathExpander(AbstractPathExpander):
    """
    Class for expanding file paths and extracting metadata on a local filesystem.

    See https://neuroconv.readthedocs.io/en/main/user_guide/expand_path.html for more information.
    """

    def list_directory(self, base_directory: DirectoryPath) -> Iterable[FilePath]:  # noqa: D101
        base_directory = Path(base_directory)
        assert base_directory.is_dir(), f"The specified 'base_directory' ({base_directory}) is not a directory!"
        return (str(path.relative_to(base_directory)) for path in base_directory.rglob("*"))


def construct_path_template(path: str, *, subject_id: str, session_id: str, **metadata_kwargs) -> str:
    """
    Construct a path template by replacing specific parts of a given path with placeholders.

    This function takes a real path example and replaces the occurrences of subject ID, session ID,
    and any additional metadata values with their respective placeholders.

    Parameters
    ----------
    path : str
        The path string containing actual data that needs to be templated.
    subject_id : str
        The subject ID in the path that will be replaced with the '{subject_id}' placeholder.
    session_id : str
        The session ID in the path that will be replaced with the '{session_id}' placeholder.
    **metadata_kwargs : dict
        Additional key-value pairs where the key is the placeholder name and the value is the actual data
        in the path that should be replaced by the placeholder.

    Returns
    -------
    str
        The path string with specified parts replaced by placeholders.

    Raises
    ------
    ValueError
        If `subject_id`, `session_id`, or any value in `metadata_kwargs` is an empty string, or if `subject_id`
        or `session_id` placeholders are not found in the path.

    Examples
    --------
    >>> construct_path_template(
    >>>     "/data/subject456/session123/file.txt",
    >>>     subject_id="subject456",
    >>>     session_id="session123"
    >>> )
    '/data/{subject_id}/{session_id}/file.txt'

    >>> construct_path_template(
    >>>     "/data/subject789/session456/image.txt",
    >>>     subject_id="subject789",
    >>>     session_id="session456",
    >>>     file_type="image"
    >>> )
    '/data/{subject_id}/{session_id}/{file_type}.txt'
    """

    if subject_id == "" or session_id == "":
        raise ValueError("Subject ID and Session ID cannot be empty strings")

    if subject_id not in path or session_id not in path:
        raise ValueError("Subject ID and Session ID must be present in the path")

    format_string = path.replace(subject_id, "{subject_id}").replace(session_id, "{session_id}")

    for key, val in metadata_kwargs.items():
        if val == "":
            raise ValueError(f"Value for '{key}' cannot be an empty string")
        if val not in format_string:
            raise ValueError(f"Value for '{key}' not found in the path")
        format_string = format_string.replace(val, f"{{{key}}}")

    return format_string



import collections.abc
import inspect
import json
import warnings
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, Optional

import docstring_parser
import hdmf.data_utils
import numpy as np
import pydantic
import pynwb
from jsonschema import validate
from pynwb.device import Device
from pynwb.icephys import IntracellularElectrode


class _NWBMetaDataEncoder(json.JSONEncoder):
    """
    Custom JSON encoder for NWB metadata.

    This encoder extends the default JSONEncoder class and provides custom serialization
    for certain data types commonly used in NWB metadata.
    """

    def default(self, obj):
        """
        Serialize custom data types to JSON. This overwrites the default method of the JSONEncoder class.
        """
        # Over-write behaviors for datetime object
        if isinstance(obj, datetime):
            return obj.isoformat()

        # Transform numpy generic integers and floats to python ints floats
        if isinstance(obj, np.generic):
            return obj.item()

        if isinstance(obj, np.ndarray):
            return obj.tolist()

        # The base-class handles it
        return super().default(obj)


class _NWBSourceDataEncoder(_NWBMetaDataEncoder):
    """
    Custom JSON encoder for data interface source data (i.e. kwargs).

    This encoder extends the default JSONEncoder class and provides custom serialization
    for certain data types commonly used in interface source data.
    """

    def default(self, obj):

        # Over-write behaviors for Paths
        if isinstance(obj, Path):
            return str(obj)

        return super().default(obj)


def get_base_schema(
    tag: Optional[str] = None,
    root: bool = False,
    id_: Optional[str] = None,
    required: Optional[list[str]] = None,
    properties: Optional[dict] = None,
    **kwargs,
) -> dict:
    """Return the base schema used for all other schemas."""
    base_schema = dict(
        required=required or [],
        properties=properties or {},
        type="object",
        additionalProperties=False,
    )
    if tag is not None:
        base_schema.update(tag=tag)
    if root:
        base_schema.update({"$schema": "http://json-schema.org/draft-07/schema#"})
    if id_ is not None:
        base_schema.update({"$id": id_})
    base_schema.update(**kwargs)
    return base_schema


def get_schema_from_method_signature(method: Callable, exclude: Optional[list[str]] = None) -> dict:
    """Deprecated version of `get_json_schema_from_method_signature`."""
    message = (
        "The method `get_schema_from_method_signature` is now named `get_json_schema_from_method_signature`."
        "This method is deprecated and will be removed after January 2025."
    )
    warnings.warn(message=message, category=DeprecationWarning, stacklevel=2)

    return get_json_schema_from_method_signature(method=method, exclude=exclude)


def get_json_schema_from_method_signature(method: Callable, exclude: Optional[list[str]] = None) -> dict:
    """
    Get the equivalent JSON schema for a signature of a method.

    Also uses `docstring_parser` (NumPy style) to attempt to find descriptions for the arguments.

    Parameters
    ----------
    method : callable
        The method to generate the JSON schema from.
    exclude : list of str, optional
        List of arguments to exclude from the schema generation.
        Always includes 'self' and 'cls'.

    Returns
    -------
    json_schema : dict
        The JSON schema corresponding to the method signature.
    """
    exclude = exclude or []
    exclude += ["self", "cls"]

    split_qualname = method.__qualname__.split(".")[-2:]
    method_display = ".".join(split_qualname) if "<" not in split_qualname[0] else method.__name__

    signature = inspect.signature(obj=method)
    parameters = signature.parameters
    additional_properties = False
    arguments_to_annotations = {}
    for argument_name in parameters:
        if argument_name in exclude:
            continue

        parameter = parameters[argument_name]

        if parameter.kind == inspect.Parameter.VAR_KEYWORD:  # Skip all **{...} usage
            additional_properties = True
            continue

        annotation = parameter.annotation

        # Pydantic uses ellipsis for required
        pydantic_default = ... if parameter.default is inspect._empty else parameter.default

        arguments_to_annotations.update({argument_name: (annotation, pydantic_default)})

    # The ConfigDict is required to support custom types like NumPy arrays
    model = pydantic.create_model(
        "_TempModel", __config__=pydantic.ConfigDict(arbitrary_types_allowed=True), **arguments_to_annotations
    )

    temp_json_schema = model.model_json_schema()

    # We never used to include titles in the lower schema layers
    # But Pydantic does automatically
    json_schema = _copy_without_title_keys(temp_json_schema)

    # Pydantic does not make determinations on additionalProperties
    json_schema["additionalProperties"] = additional_properties

    # Attempt to find descriptions within the docstring of the method
    parsed_docstring = docstring_parser.parse(method.__doc__)
    for parameter_in_docstring in parsed_docstring.params:
        if parameter_in_docstring.arg_name in exclude:
            continue

        if parameter_in_docstring.arg_name not in json_schema["properties"]:
            message = (
                f"The argument_name '{parameter_in_docstring.arg_name}' from the docstring of method "
                f"'{method_display}' does not occur in the signature, possibly due to a typo."
            )
            warnings.warn(message=message, stacklevel=2)
            continue

        if parameter_in_docstring.description is not None:
            json_schema["properties"][parameter_in_docstring.arg_name].update(
                description=parameter_in_docstring.description
            )
    # TODO: could also add Field support for more direct control over docstrings (and enhanced validation conditions)

    return json_schema


def _copy_without_title_keys(d: Any, /) -> Optional[dict]:
    if not isinstance(d, dict):
        return d

    return {key: _copy_without_title_keys(value) for key, value in d.items() if key != "title"}


def fill_defaults(schema: dict, defaults: dict, overwrite: bool = True):
    """
    Insert the values of the defaults dict as default values in the schema in place.

    Parameters
    ----------
    schema: dict
    defaults: dict
    overwrite: bool
    """
    # patternProperties introduced with the CsvTimeIntervalsInterface
    # caused issue with NWBConverter.get_metadata_schema() call leading here
    properties_reference = "properties"
    if properties_reference not in schema and "patternProperties" in schema:
        properties_reference = "patternProperties"

    for key, val in schema[properties_reference].items():
        if key in defaults:
            if val["type"] == "object":
                fill_defaults(val, defaults[key], overwrite=overwrite)
            else:
                if overwrite or ("default" not in val):
                    val["default"] = defaults[key]


def unroot_schema(schema: dict):
    """
    Modify a json-schema dictionary to make it not root.

    Parameters
    ----------
    schema: dict
    """
    terms = ("required", "properties", "type", "additionalProperties", "title", "description")
    return {k: v for k, v in schema.items() if k in terms}


def _is_member(types, target_types):
    if not isinstance(target_types, tuple):
        target_types = (target_types,)
    if not isinstance(types, tuple):
        types = (types,)
    return any(t in target_types for t in types)


def get_schema_from_hdmf_class(hdmf_class):
    """Get metadata schema from hdmf class."""
    schema = get_base_schema()
    schema["tag"] = hdmf_class.__module__ + "." + hdmf_class.__name__

    # Detect child-like (as opposed to link) fields
    pynwb_children_fields = [f["name"] for f in hdmf_class.get_fields_conf() if f.get("child", False)]
    # For MultiContainerInterface
    if hasattr(hdmf_class, "__clsconf__"):
        pynwb_children_fields.append(hdmf_class.__clsconf__["attr"])
    # Temporary solution before this is solved: https://github.com/hdmf-dev/hdmf/issues/475
    if "device" in pynwb_children_fields:
        pynwb_children_fields.remove("device")
    docval = hdmf_class.__init__.__docval__
    for docval_arg in docval["args"]:
        arg_name = docval_arg["name"]
        arg_type = docval_arg["type"]

        schema_val = dict(description=docval_arg["doc"])

        if arg_name == "name":
            schema_val.update(pattern="^[^/]*$")

        if _is_member(arg_type, (float, int, "float", "int")):
            schema_val.update(type="number")
        elif _is_member(arg_type, str):
            schema_val.update(type="string")
        elif _is_member(arg_type, collections.abc.Iterable):
            schema_val.update(type="array")
        elif isinstance(arg_type, tuple) and (np.ndarray in arg_type and hdmf.data_utils.DataIO not in arg_type):
            # extend type array without including type where DataIO in tuple
            schema_val.update(type="array")
        elif _is_member(arg_type, datetime):
            schema_val.update(type="string", format="date-time")
        elif _is_member(arg_type, (pynwb.base.TimeSeries, pynwb.ophys.PlaneSegmentation)):
            continue
        else:
            if not isinstance(arg_type, tuple):
                docval_arg_type = [arg_type]
            else:
                docval_arg_type = arg_type
            # if another nwb object (or list of nwb objects)
            if any([hasattr(t, "__nwbfields__") for t in docval_arg_type]):
                is_nwb = [hasattr(t, "__nwbfields__") for t in docval_arg_type]
                item = docval_arg_type[np.where(is_nwb)[0][0]]
                # if it is child
                if arg_name in pynwb_children_fields:
                    items = get_schema_from_hdmf_class(item)
                    schema_val.update(type="array", items=items, minItems=1, maxItems=1)
                # if it is a link
                else:
                    target = item.__module__ + "." + item.__name__
                    schema_val.update(type="string", target=target)
            else:
                continue
        # Check for default arguments
        if "default" in docval_arg:
            if docval_arg["default"] is not None:
                schema_val.update(default=docval_arg["default"])
        else:
            schema["required"].append(arg_name)
        schema["properties"][arg_name] = schema_val
    if "allow_extra" in docval:
        schema["additionalProperties"] = docval["allow_extra"]
    return schema


def get_metadata_schema_for_icephys() -> dict:
    """
    Returns the metadata schema for icephys data.

    Returns:
        dict: The metadata schema for icephys data.

    """
    schema = get_base_schema(tag="Icephys")
    schema["required"] = ["Device", "Electrodes"]
    schema["properties"] = dict(
        Device=dict(type="array", minItems=1, items={"$ref": "#/properties/Icephys/definitions/Device"}),
        Electrodes=dict(
            type="array",
            minItems=1,
            items={"$ref": "#/properties/Icephys/definitions/Electrode"},
        ),
        Sessions=dict(
            type="array",
            minItems=1,
            items={"$ref": "#/properties/Icephys/definitions/Sessions"},
        ),
    )

    schema["definitions"] = dict(
        Device=get_schema_from_hdmf_class(Device),
        Electrode=get_schema_from_hdmf_class(IntracellularElectrode),
        Sessions=dict(
            name={"type": "string", "description": "Session name."},
            relative_session_start_time={
                "type": "number",
                "description": "the start time of the sessions in seconds, relative to the absolute start time",
            },
            icephys_experiment_type={
                "type": "string",
                "description": "Icephys experiment type. Allowed types are: voltage_clamp, current_clamp and izero",
            },
            stimulus_type={
                "type": "string",
                "description": "Description of the type pf stimulus, e.g. Square current clamp.",
            },
            recordings=dict(
                type="array",
                minItems=1,
                items={"$ref": "#/properties/Icephys/definitions/SessionsRecordings"},
            ),
        ),
        SessionsRecordings=dict(
            intracellular_recordings_table_ind={"type": "number", "description": ""},
            simultaneous_recordings_table_ind={"type": "number", "description": ""},
            sequential_recordings_table_ind={"type": "number", "description": ""},
        ),
    )

    return schema


def validate_metadata(metadata: dict[str, dict], schema: dict[str, dict], verbose: bool = False):
    """Validate metadata against a schema."""
    encoder = _NWBMetaDataEncoder()
    # The encoder produces a serialized object, so we deserialized it for comparison

    serialized_metadata = encoder.encode(metadata)
    decoded_metadata = json.loads(serialized_metadata)
    validate(instance=decoded_metadata, schema=schema)
    if verbose:
        print("Metadata is valid!")



import math


def human_readable_size(size_bytes: int, binary: bool = False) -> str:
    """
    Convert a file size given in bytes to a human-readable format using division
    and remainder instead of iteration.

    Parameters
    ----------
    size_bytes : int
        The size in bytes.
    binary : bool, default=False
        If True, use binary prefixes (KiB, MiB, etc.). If False, use SI prefixes (KB, MB, etc.).

    Returns
    -------
    str
        A human-readable string representation of the size.

    Examples
    --------
    >>> human_readable_size(123)
    '123 B'
    >>> human_readable_size(1234, binary=True)
    '1.21 KiB'
    >>> human_readable_size(123456789)
    '123.46 MB'
    """
    # Check if size is negative
    if size_bytes < 0:
        raise ValueError("Size must be non-negative")

    if size_bytes == 0:
        return "0 B"

    # Define the suffixes for each size unit
    suffixes = ["", "K", "M", "G", "T", "P", "E", "Z", "Y"]

    # Calculate base and the exponent
    base = 1024 if binary else 1000
    exponent = int(math.log(size_bytes, base))

    if exponent == 0:
        return f"{size_bytes} B"

    # Calculate the human-readable size
    human_readable_value = size_bytes / (base**exponent)

    # Return formatted size with suffix
    return f"{human_readable_value:.2f} {suffixes[exponent]}{'i' if binary else ''}B"



from pathlib import Path
from typing import Optional

from pydantic import FilePath

from .header_tools import _parse_nev_basic_header, _parse_nsx_basic_header
from ..baserecordingextractorinterface import BaseRecordingExtractorInterface
from ..basesortingextractorinterface import BaseSortingExtractorInterface
from ....utils import get_schema_from_method_signature


class BlackrockRecordingInterface(BaseRecordingExtractorInterface):
    """Primary data interface class for converting Blackrock data using a
    :py:class:`~spikeinterface.extractors.BlackrockRecordingExtractor`."""

    display_name = "Blackrock Recording"
    associated_suffixes = (".ns0", ".ns1", ".ns2", ".ns3", ".ns4", ".ns5")
    info = "Interface for Blackrock recording data."

    @classmethod
    def get_source_schema(cls):
        source_schema = get_schema_from_method_signature(method=cls.__init__, exclude=["block_index", "seg_index"])
        source_schema["properties"]["file_path"][
            "description"
        ] = "Path to the Blackrock file with suffix being .ns1, .ns2, .ns3, .ns4m .ns4, or .ns6."
        return source_schema

    def __init__(
        self,
        file_path: FilePath,
        nsx_override: Optional[FilePath] = None,
        verbose: bool = True,
        es_key: str = "ElectricalSeries",
    ):
        """
        Load and prepare data corresponding to Blackrock interface.

        Parameters
        ----------
        file_path : FilePathType
            Path to the Blackrock file with suffix being .ns1, .ns2, .ns3, .ns4m .ns4, or .ns6
        verbose: bool, default: True
        es_key : str, default: "ElectricalSeries"
        """

        file_path = Path(file_path)
        if file_path.suffix == "":
            assert nsx_override is not None, (
                "if file_path is empty " 'provide a nsx file to load with "nsx_override" arg'
            )
            nsx_to_load = None
            self.file_path = Path(nsx_override)
        else:
            assert "ns" in file_path.suffix, "file_path should be an nsx file"
            nsx_to_load = int(file_path.suffix[-1])
            self.file_path = file_path

        super().__init__(file_path=file_path, stream_id=str(nsx_to_load), verbose=verbose, es_key=es_key)

    def get_metadata(self) -> dict:
        metadata = super().get_metadata()
        # Open file and extract headers
        basic_header = _parse_nsx_basic_header(self.source_data["file_path"])
        if "TimeOrigin" in basic_header:
            metadata["NWBFile"].update(session_start_time=basic_header["TimeOrigin"])
        if "Comment" in basic_header:
            metadata["NWBFile"].update(session_description=basic_header["Comment"])

        return metadata


class BlackrockSortingInterface(BaseSortingExtractorInterface):
    """Primary data interface class for converting Blackrock spiking data."""

    display_name = "Blackrock Sorting"
    associated_suffixes = (".nev",)
    info = "Interface for Blackrock sorting data."

    @classmethod
    def get_source_schema(cls) -> dict:
        metadata_schema = get_schema_from_method_signature(method=cls.__init__)
        metadata_schema["additionalProperties"] = True
        metadata_schema["properties"]["file_path"].update(description="Path to Blackrock .nev file.")
        return metadata_schema

    def __init__(self, file_path: FilePath, sampling_frequency: float = None, verbose: bool = True):
        """
        Parameters
        ----------
        file_path : str, Path
            The file path to the ``.nev`` data
        sampling_frequency: float, optional
            The sampling frequency for the sorting extractor. When the signal data is available (.ncs) those files will be
        used to extract the frequency automatically. Otherwise, the sampling frequency needs to be specified for
        this extractor to be initialized.
        verbose : bool, default: True
            Enables verbosity
        """
        super().__init__(file_path=file_path, sampling_frequency=sampling_frequency, verbose=verbose)

    def get_metadata(self) -> dict:
        metadata = super().get_metadata()
        # Open file and extract headers
        basic_header = _parse_nev_basic_header(self.source_data["file_path"])
        if "TimeOrigin" in basic_header:
            session_start_time = basic_header["TimeOrigin"]
            metadata["NWBFile"].update(session_start_time=session_start_time.strftime("%Y-%m-%dT%H:%M:%S"))
        if "Comment" in basic_header:
            metadata["NWBFile"].update(session_description=basic_header["Comment"])
        return metadata



"""Base Pydantic models for DatasetInfo and DatasetConfiguration."""

from typing import Any, ClassVar, Literal, Type

from hdmf.container import DataIO
from pydantic import BaseModel, ConfigDict, Field
from pynwb import NWBFile
from typing_extensions import Self

from ._base_dataset_io import DatasetIOConfiguration
from ._pydantic_pure_json_schema_generator import PureJSONSchemaGenerator
from .._dataset_configuration import get_default_dataset_io_configurations


class BackendConfiguration(BaseModel):
    """A model for matching collections of DatasetConfigurations to a specific backend."""

    backend: ClassVar[Literal["hdf5", "zarr"]]
    pretty_backend_name: ClassVar[Literal["HDF5", "Zarr"]]
    data_io_class: ClassVar[Type[DataIO]]

    model_config = ConfigDict(validate_assignment=True)  # Re-validate model on mutation

    dataset_configurations: dict[str, DatasetIOConfiguration] = Field(
        description=(
            "A mapping from object locations (e.g. `acquisition/TestElectricalSeriesAP/data`) "
            "to their DatasetConfiguration specification that contains all information "
            "for writing the datasets to disk using the specific backend."
        )
    )

    def __str__(self) -> str:
        """Not overriding __repr__ as this is intended to render only when wrapped in print()."""
        string = (
            f"\n{self.pretty_backend_name} dataset configurations" f"\n{'-' * (len(self.pretty_backend_name) + 23)}"
        )

        for dataset_configuration in self.dataset_configurations.values():
            string += f"\n{dataset_configuration}"

        return string

    # Pydantic models have several API calls for retrieving the schema - override all of them to work
    @classmethod
    def schema(cls, **kwargs) -> dict[str, Any]:
        return cls.model_json_schema(**kwargs)

    @classmethod
    def schema_json(cls, **kwargs) -> dict[str, Any]:
        return cls.model_json_schema(**kwargs)

    @classmethod
    def model_json_schema(cls, **kwargs) -> dict[str, Any]:
        assert "mode" not in kwargs, "The 'mode' of this method is fixed to be 'validation' and cannot be changed."
        assert "schema_generator" not in kwargs, "The 'schema_generator' of this method cannot be changed."
        return super().model_json_schema(mode="validation", schema_generator=PureJSONSchemaGenerator, **kwargs)

    @classmethod
    def from_nwbfile(cls, nwbfile: NWBFile) -> Self:
        default_dataset_configurations = get_default_dataset_io_configurations(nwbfile=nwbfile, backend=cls.backend)
        dataset_configurations = {
            default_dataset_configuration.location_in_file: default_dataset_configuration
            for default_dataset_configuration in default_dataset_configurations
        }

        return cls(dataset_configurations=dataset_configurations)

    def find_locations_requiring_remapping(self, nwbfile: NWBFile) -> dict[str, DatasetIOConfiguration]:
        """
        Find locations of objects with mismatched IDs in the file.

        This function identifies neurodata objects in the `nwbfile` that have matching locations
        with the current configuration but different object IDs. It returns a dictionary of
        remapped `DatasetIOConfiguration` objects for these mismatched locations.

        Parameters
        ----------
        nwbfile : pynwb.NWBFile
            The NWBFile object to check for mismatched object IDs.

        Returns
        -------
        dict[str, DatasetIOConfiguration]
            A dictionary where:
            * Keys: Locations in the NWB of objects with mismatched IDs.
            * Values: New `DatasetIOConfiguration` objects corresponding to the updated object IDs.

        Notes
        -----
        * This function only checks for objects with the same location but different IDs.
        * It does not identify objects missing from the current configuration.
        * The returned `DatasetIOConfiguration` objects are copies of the original configurations
        with updated `object_id` fields.
        """
        # Use a fresh default configuration to get mapping of object IDs to locations in file
        default_configurations = list(get_default_dataset_io_configurations(nwbfile=nwbfile, backend=self.backend))

        if len(default_configurations) != len(self.dataset_configurations):
            raise ValueError(
                f"The number of default configurations ({len(default_configurations)}) does not match the number of "
                f"specified configurations ({len(self.dataset_configurations)})!"
            )

        objects_requiring_remapping = {}
        for dataset_configuration in default_configurations:
            location_in_file = dataset_configuration.location_in_file
            object_id = dataset_configuration.object_id

            location_cannot_be_remapped = location_in_file not in self.dataset_configurations
            if location_cannot_be_remapped:
                raise KeyError(
                    f"Unable to remap the object IDs for object at location '{location_in_file}'! This "
                    "usually occurs if you are attempting to configure the backend for two files of "
                    "non-equivalent structure."
                )

            former_configuration = self.dataset_configurations[location_in_file]
            former_object_id = former_configuration.object_id

            if former_object_id == object_id:
                continue

            remapped_configuration = former_configuration.model_copy(update={"object_id": object_id})
            objects_requiring_remapping[location_in_file] = remapped_configuration

        return objects_requiring_remapping

    def build_remapped_backend(
        self,
        locations_to_remap: dict[str, DatasetIOConfiguration],
    ) -> Self:
        """
        Build a remapped backend configuration by updating mismatched object IDs.

        This function takes a dictionary of new `DatasetIOConfiguration` objects
        (as returned by `find_locations_requiring_remapping`) and updates a copy of the current configuration
        with these new configurations.

        Parameters
        ----------
        locations_to_remap : dict
            A dictionary mapping locations in the NWBFile to their corresponding new
            `DatasetIOConfiguration` objects with updated IDs.

        Returns
        -------
        Self
            A new instance of the backend configuration class with updated object IDs for
            the specified locations.
        """
        new_backend_configuration = self.model_copy(deep=True)
        new_backend_configuration.dataset_configurations.update(locations_to_remap)
        return new_backend_configuration



from typing import Literal

from pydantic import FilePath, validate_call

from ..baseimagingextractorinterface import BaseImagingExtractorInterface


class TiffImagingInterface(BaseImagingExtractorInterface):
    """Interface for multi-page TIFF files."""

    display_name = "TIFF Imaging"
    associated_suffixes = (".tif", ".tiff")
    info = "Interface for multi-page TIFF files."

    @classmethod
    def get_source_schema(cls) -> dict:
        source_schema = super().get_source_schema()
        source_schema["properties"]["file_path"]["description"] = "Path to Tiff file."
        return source_schema

    @validate_call
    def __init__(
        self,
        file_path: FilePath,
        sampling_frequency: float,
        verbose: bool = True,
        photon_series_type: Literal["OnePhotonSeries", "TwoPhotonSeries"] = "TwoPhotonSeries",
    ):
        """
        Initialize reading of TIFF file.

        Parameters
        ----------
        file_path : FilePathType
        sampling_frequency : float
        verbose : bool, default: True
        photon_series_type : {'OnePhotonSeries', 'TwoPhotonSeries'}, default: "TwoPhotonSeries"
        """
        super().__init__(
            file_path=file_path,
            sampling_frequency=sampling_frequency,
            verbose=verbose,
            photon_series_type=photon_series_type,
        )



from dateutil.parser import parse
from pydantic import DirectoryPath, validate_call

from ..baseimagingextractorinterface import BaseImagingExtractorInterface


class MicroManagerTiffImagingInterface(BaseImagingExtractorInterface):
    """Data Interface for MicroManagerTiffImagingExtractor."""

    display_name = "Micro-Manager TIFF Imaging"
    associated_suffixes = (".ome", ".tif", ".json")
    info = "Interface for Micro-Manager TIFF imaging data."

    @classmethod
    def get_source_schema(cls) -> dict:
        source_schema = super().get_source_schema()

        source_schema["properties"]["folder_path"]["description"] = "The folder containing the OME-TIF image files."
        return source_schema

    @validate_call
    def __init__(self, folder_path: DirectoryPath, verbose: bool = True):
        """
        Data Interface for MicroManagerTiffImagingExtractor.

        Parameters
        ----------
        folder_path : FolderPathType
            The folder path that contains the OME-TIF image files (.ome.tif files) and
           the 'DisplaySettings' JSON file.
        verbose : bool, default: True
        """
        super().__init__(folder_path=folder_path)
        self.verbose = verbose
        # Micro-Manager uses "Default" as channel name, for clarity we rename it to  'OpticalChannelDefault'
        channel_name = self.imaging_extractor._channel_names[0]
        self.imaging_extractor._channel_names = [f"OpticalChannel{channel_name}"]

    def get_metadata(self) -> dict:
        metadata = super().get_metadata()

        micromanager_metadata = self.imaging_extractor.micromanager_metadata
        session_start_time = parse(micromanager_metadata["Summary"]["StartTime"])
        metadata["NWBFile"].update(session_start_time=session_start_time)

        imaging_plane_metadata = metadata["Ophys"]["ImagingPlane"][0]
        imaging_plane_metadata.update(
            imaging_rate=self.imaging_extractor.get_sampling_frequency(),
        )
        metadata["Ophys"]["TwoPhotonSeries"][0].update(
            unit="px",
            format="tiff",
        )

        return metadata



import numpy as np


def create_optogenetic_stimulation_timeseries(
    *,
    stimulation_onset_times: np.ndarray,
    duration: float,
    frequency: float,
    pulse_width: float,
    power: float,
) -> tuple[np.ndarray, np.ndarray]:
    """Create a continuous stimulation time series from stimulation onset times and parameters.

    In the resulting data array, the offset time of each pulse is represented by a 0 power value.

    Parameters
    ----------
    stimulation_onset_times : np.ndarray
        Array of stimulation onset times.
    duration : float
        Duration of stimulation in seconds.
    frequency : float
        Frequency of stimulation in Hz.
    pulse_width : float
        Pulse width of stimulation in seconds.
    power : float
        Power of stimulation in W.

    Returns
    -------
    np.ndarray
        Stimulation timestamps.
    np.ndarray
        Instantaneous stimulation power.

    Notes
    -----
    For continuous (non-pulsed) stimulation of a desired duration, simply set
    ```
    pulse_width = duration
    frequency = 1 / duration
    ```
    """
    num_pulses = int(duration * frequency)
    inter_pulse_interval = 1 / frequency
    timestamps, data = [0], [0]
    for onset_time in stimulation_onset_times:
        for i in range(num_pulses):
            pulse_onset_time = onset_time + i * inter_pulse_interval
            timestamps.append(pulse_onset_time)
            data.append(power)
            pulse_offset_time = pulse_onset_time + pulse_width
            timestamps.append(pulse_offset_time)
            data.append(0)
    return np.array(timestamps, dtype=np.float64), np.array(data, dtype=np.float64)
